#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import pytest
from construct import ValidationError

from bluetooth_mesh.messages.silvair.network_diagnostic_server import (
    NetworkDiagnosticServerParams,
    NetworkDiagnosticServerSubOpcode,
    NetworkDiagnosticSetupServerParams,
    NetworkDiagnosticSetupServerSubOpcode,
)

# fmt: off
valid = [
    pytest.param(
        NetworkDiagnosticSetupServerParams,
        bytes.fromhex("00"),
        NetworkDiagnosticSetupServerSubOpcode.PUBLICATION_GET,
        {},
        id='PublicationGet'
    ),
    pytest.param(
        NetworkDiagnosticSetupServerParams,
        bytes.fromhex("01 1234 AAAA 82 03 1000"),
        NetworkDiagnosticSetupServerSubOpcode.PUBLICATION_SET,
        dict(
            destination=0x3412,
            count=0xAAAA,
            period=dict(steps=0x02, resolution=0b10),
            ttl=0x03,
            net_key_index=0x0010,
            features=None
        ),
        id='PublicationSet (no features)'
    ),
    pytest.param(
        NetworkDiagnosticSetupServerParams,
        bytes.fromhex("01 1234 AAAA 82 03 1000 0300"),
        NetworkDiagnosticSetupServerSubOpcode.PUBLICATION_SET,
        dict(
            destination=0x3412,
            count=0xAAAA,
            period=dict(steps=0x02, resolution=0b10),
            ttl=0x03,
            net_key_index=0x0010,
            features=0x0003
        ),
        id="PublicationSet (with features)"
    ),
    pytest.param(
        NetworkDiagnosticSetupServerParams,
        bytes.fromhex("02 1234 AAAA 82 03 1000"),
        NetworkDiagnosticSetupServerSubOpcode.PUBLICATION_STATUS,
        dict(
            destination=0x3412,
            count=0xAAAA,
            period=dict(steps=0x02, resolution=0b10),
            ttl=0x03,
            net_key_index=0x0010,
            features=None
        ),
        id="PublicationStatus (no features)"
    ),
    pytest.param(
        NetworkDiagnosticSetupServerParams,
        bytes.fromhex("02 1234 AAAA 82 03 1000 0300"),
        NetworkDiagnosticSetupServerSubOpcode.PUBLICATION_STATUS,
        dict(
            destination=0x3412,
            count=0xAAAA,
            period=dict(steps=0x02, resolution=0b10),
            ttl=0x03,
            net_key_index=0x0010,
            features=0x0003
        ),
        id="PublicationStatus (with features)"
    ),
    pytest.param(
        NetworkDiagnosticServerParams,
        bytes.fromhex("00"),
        NetworkDiagnosticServerSubOpcode.SUBSCRIPTION_GET,
        {},
        id="SubscriptionGet"
    ),
    pytest.param(
        NetworkDiagnosticServerParams,
        bytes.fromhex("01 1234 AAAA"),
        NetworkDiagnosticServerSubOpcode.SUBSCRIPTION_SET,
        dict(
            destination=0x3412,
            period=0xAAAA
        ),
        id="SubscriptionSet"
    ),
    pytest.param(
        NetworkDiagnosticServerParams,
        bytes.fromhex("03 1234 AAAA 20 5678 BBBB 00 00"),
        NetworkDiagnosticServerSubOpcode.SUBSCRIPTION_STATUS,
        dict(
            destination=0x3412,
            period=0xAAAA,
            max_record_count=32,
            record=[dict(source=0x7856, count=0xBBBB, min_hops=0x00, max_hops=0x00)]
        ),
        id="SubscriptionStatus (one record)"
    ),
    pytest.param(
        NetworkDiagnosticServerParams,
        bytes.fromhex("03 1234 AAAA 20 5678 BBBB 00 00 9101 CCCC 01 01"),
        NetworkDiagnosticServerSubOpcode.SUBSCRIPTION_STATUS,
        dict(
            destination=0x3412,
            period=0xAAAA,
            max_record_count=32,
            record=[
                dict(source=0x7856, count=0xBBBB, min_hops=0x00, max_hops=0x00),
                dict(source=0x0191, count=0xCCCC, min_hops=0x01, max_hops=0x01),
            ]
        ),
        id="SubscriptionStatus (multiple records)"
    )
]
# fmt: on


@pytest.mark.parametrize("klass,encoded,subopcode,payload", valid)
def test_parse_valid(klass, encoded, subopcode, payload):
    assert klass.parse(encoded).subopcode == subopcode
    assert klass.parse(encoded).payload == payload


@pytest.mark.parametrize("klass,encoded,subopcode,payload", valid)
def test_build_valid(klass, encoded, subopcode, payload):
    assert klass.build(dict(subopcode=subopcode, payload=payload)) == encoded


def test_prohibited_destination_addr_publication_set():
    with pytest.raises(ValidationError):
        NetworkDiagnosticServerParams.build(
            dict(
                subopcode=NetworkDiagnosticSetupServerSubOpcode.PUBLICATION_SET,
                payload=dict(
                    destination=0xFF00,
                    count=0xAAAA,
                    period=dict(steps=0x02, resolution=0b10),
                    ttl=0x03,
                    net_key_index=0x0010,
                    features=None,
                ),
            )
        )


def test_prohibited_destination_addr_subscription_set():
    with pytest.raises(ValidationError):
        NetworkDiagnosticServerParams.build(
            dict(
                subopcode=NetworkDiagnosticServerSubOpcode.SUBSCRIPTION_SET,
                payload=dict(destination=0xFF00, period=0xAAAA),
            )
        )


def test_prohibited_source_addr_subscription_stat():
    with pytest.raises(ValidationError):
        NetworkDiagnosticServerParams.build(
            dict(
                subopcode=NetworkDiagnosticServerSubOpcode.SUBSCRIPTION_STATUS,
                payload=dict(
                    destination=0x3412,
                    period=0xAAAA,
                    max_record_count=32,
                    record=[dict(source=0xFF00, count=0xBBBB, min_hops=0x00, max_hops=0x00)],
                ),
            )
        )



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from enum import IntEnum

from construct import BitsInteger, BitStruct, Int8ul, Int24ul, Struct, this

from bluetooth_mesh.messages.util import (
    DefaultCountValidator,
    EnumAdapter,
    EnumSwitch as Switch,
    Opcode,
    SwitchStruct,
)


class GenericBatteryOpcode(IntEnum):
    GENERIC_BATTERY_GET = 0x8223
    GENERIC_BATTERY_STATUS = 0x8224


class GenericBatteryFlagsPresence(IntEnum):
    BATTERY_NOT_PRESENT = 0b00
    BATTERY_PRESENT_REMOVABLE = 0b01
    BATTERY_PRESENT_NON_REMOVABLE = 0b10
    BATTERY_PRESENCE_UNKNOWN = 0b11


class GenericBatteryFlagsIndicator(IntEnum):
    BATTERY_CHARGE_CRITICALLY_LOW = 0b00
    BATTERY_CHARGE_LOW = 0b01
    BATTERY_CHARGE_GOOD = 0b10
    BATTERY_CHARGE_UNKNOWN = 0b11


class GenericBatteryFlagsCharging(IntEnum):
    BATTERY_NOT_CHARGEABLE = 0b00
    BATTERY_CHARGEABLE_NOT_CHARGING = 0b01
    BATTERY_CHARGEABLE_CHARGING = 0b10
    BATTERY_CHARGING_STATE_UNKNOWN = 0b11


class GenericBatteryFlagsServiceability(IntEnum):
    RFU = 0b00
    BATTERY_NOT_REQUIRE_SERVICE = 0b01
    BATTERY_REQUIRE_SERVICE = 0b10
    BATTERY_SERVICEABILITY_UNKNOWN = 0b11


# fmt: off
BatteryFlags = BitStruct(
    "battery_serviceability_flags" / EnumAdapter(BitsInteger(2),
                                                 GenericBatteryFlagsServiceability),
    "battery_charging_flags" / EnumAdapter(BitsInteger(2),
                                           GenericBatteryFlagsCharging),
    "battery_indicator_flags" / EnumAdapter(BitsInteger(2),
                                            GenericBatteryFlagsIndicator),
    "battery_presence_flags" / EnumAdapter(BitsInteger(2),
                                           GenericBatteryFlagsPresence),
)

GenericBatteryGet = Struct()

GenericBatteryStatus = Struct(
    "battery_level" / DefaultCountValidator(Int8ul),
    "time_to_discharge" / DefaultCountValidator(Int24ul),
    "time_to_charge" / DefaultCountValidator(Int24ul),
    "flags" / BatteryFlags
)

GenericBatteryMessage = SwitchStruct(
    "opcode" / Opcode(GenericBatteryOpcode),
    "params" / Switch(
        this.opcode,
        {
            GenericBatteryOpcode.GENERIC_BATTERY_GET: GenericBatteryGet,
            GenericBatteryOpcode.GENERIC_BATTERY_STATUS: GenericBatteryStatus,
        },
    )
)
# fmt: on



from .ctl import LightCTLMessage, LightCTLOpcode, LightCTLSetupMessage, LightCTLSetupOpcode
from .lightness import (
    LightLightnessMessage,
    LightLightnessOpcode,
    LightLightnessSetupMessage,
    LightLightnessSetupOpcode,
)



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2023  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from datetime import datetime, timedelta, timezone
from enum import IntEnum

from construct import (
    Adapter,
    BitsInteger,
    BitStruct,
    BytesInteger,
    Const,
    Container,
    Embedded,
    Flag,
    Int8ul,
    Int16sl,
    Int16ul,
    Int32ul,
    Padding,
    Select,
    Struct,
    this,
)

from bluetooth_mesh.messages.properties import PropertyDict, PropertyID
from bluetooth_mesh.messages.time import (
    MESH_UNIX_EPOCH_DIFF,
    DateTime,
    TAIUTCDeltaAdapter,
    mesh_tai_utc_delta_to_timedelta,
    mesh_time_zone_offset_to_timedelta,
    timedelta_to_mesh_tai_utc_delta,
    timedelta_to_mesh_time_zone_offset,
)
from bluetooth_mesh.messages.util import (
    EmbeddedBitStruct,
    EnumAdapter,
    EnumSwitch as Switch,
    FieldAdapter,
    NamedSelect,
    Opcode,
    SwitchStruct,
)


class EmergencyLightingTestServerOpcode(IntEnum):
    SILVAIR_ELT = 0xE93601


class EmergencyLightingTestServerSubOpcode(IntEnum):
    ELT_FUNCTIONAL_TEST_GET = 0x00
    ELT_FUNCTIONAL_TEST_START = 0x01
    ELT_FUNCTIONAL_TEST_STOP = 0x02
    ELT_FUNCTIONAL_TEST_STATUS = 0x03
    ELT_DURATION_TEST_GET = 0x04
    ELT_DURATION_TEST_START = 0x05
    ELT_DURATION_TEST_STOP = 0x06
    ELT_DURATION_TEST_STATUS = 0x07
    ELT_PROPERTY_GET = 0x08
    ELT_PROPERTY_SET = 0x09
    ELT_PROPERTY_SET_UNACKNOWLEDGED = 0x0A
    ELT_PROPERTY_STATUS = 0x0B


class TestExecutionStatus(IntEnum):
    FINISHED = 0
    TRIGGERED = 1
    IN_PROGRESS = 2
    POSTPONED = 3
    STOPPED = 4
    DROPPED = 5
    TIMEOUT = 6
    UNKNOWN = 7


FunctionalExecutionResult = BitStruct(
    Padding(5),
    "circuit_fault" / Flag,
    "battery_fault" / Flag,
    "lamp_fault" / Flag,
)

DurationExecutionResult = BitStruct(
    Padding(4),
    "battery_duration_fault" / Flag,
    "circuit_fault" / Flag,
    "battery_fault" / Flag,
    "lamp_fault" / Flag,
)


class TimestampAdapter(Adapter):
    _subcon = Struct(
        "date" / DateTime,
        "tai_utc_delta" / TAIUTCDeltaAdapter(Int16sl),
    )

    def _decode(self, obj, context, path):
        time_zone_offset = mesh_time_zone_offset_to_timedelta(obj["time_zone_offset"])
        full_recv_time = obj["tai_seconds"] + MESH_UNIX_EPOCH_DIFF + int(time_zone_offset.total_seconds())
        recv_date = datetime.fromtimestamp(full_recv_time, timezone(time_zone_offset))

        return Container(
            date=recv_date,
            tai_utc_delta=mesh_tai_utc_delta_to_timedelta(obj["tai_utc_delta"]),
        )

    def _encode(self, obj, context, path):
        passed_time: datetime = obj["date"]

        if isinstance(passed_time, dict):  # capnproto message
            time_zone = timedelta(minutes=passed_time["time_zone_offset"])
            passed_time = datetime(
                year=passed_time["year"],
                month=passed_time["month"],
                day=passed_time["day"],
                hour=passed_time["hour"],
                minute=passed_time["minute"],
                second=passed_time["second"],
                microsecond=passed_time["microsecond"],
                tzinfo=timezone(time_zone),
            )

        if isinstance(obj["tai_utc_delta"], int):
            obj["tai_utc_delta"] = timedelta(seconds=obj["tai_utc_delta"])

        total_time = passed_time.timestamp() - MESH_UNIX_EPOCH_DIFF - passed_time.utcoffset().total_seconds()

        return Container(
            tai_seconds=int(total_time),
            tai_utc_delta=timedelta_to_mesh_tai_utc_delta(obj["tai_utc_delta"]),
            time_zone_offset=timedelta_to_mesh_time_zone_offset(passed_time.utcoffset()),
        )


# fmt: off
Timestamp = Struct(
    "tai_seconds" / BytesInteger(5, swapped=True),
    "time_zone_offset" / Int8ul,
    *EmbeddedBitStruct("_",
                       Const(0, BitsInteger(1)),
                       "tai_utc_delta" / BitsInteger(15),
                       )
)

EmergencyLightingTestServerFunctionalTestGet = Struct()

EmergencyLightingTestServerFunctionalTestStart = Struct()

EmergencyLightingTestServerFunctionalTestStop = Struct()

EmergencyLightingTestServerFunctionalTestStatusMinimal = Struct(
    "tai_timestamp" / TimestampAdapter(Timestamp),
    "execution_status" / EnumAdapter(Int8ul, TestExecutionStatus),
    "execution_result" / FunctionalExecutionResult,
)

EmergencyLightingTestServerFunctionalTestStatusOptional = Struct(
    Embedded(EmergencyLightingTestServerFunctionalTestStatusMinimal),
    "relative_timestamp" / Int32ul,
)

EmergencyLightingTestServerFunctionalTestStatus = NamedSelect(
    optional=EmergencyLightingTestServerFunctionalTestStatusOptional,
    minimal=EmergencyLightingTestServerFunctionalTestStatusMinimal,
)

EmergencyLightingTestServerDurationTestGet = Struct()

EmergencyLightingTestServerDurationTestStart = Struct()

EmergencyLightingTestServerDurationTestStop = Struct()

EmergencyLightingTestServerDurationTestStatusMinimal = Struct(
    "tai_timestamp" / TimestampAdapter(Timestamp),
    "execution_status" / EnumAdapter(Int8ul, TestExecutionStatus),
    "execution_result" / DurationExecutionResult,
    "duration_result" / Int16ul,
)

EmergencyLightingTestServerDurationTestStatusOptional = Struct(
    Embedded(EmergencyLightingTestServerDurationTestStatusMinimal),
    "relative_timestamp" / Int32ul,
)

EmergencyLightingTestServerDurationTestStatus = NamedSelect(
    optional=EmergencyLightingTestServerDurationTestStatusOptional,
    minimal=EmergencyLightingTestServerDurationTestStatusMinimal,
)

EmergencyLightingTestServerPropertyId = FieldAdapter(
    Select(
        EnumAdapter(Int16ul, PropertyID),
        Int16ul
    ),
    Int16ul
)

EmergencyLightingTestServerPropertyGet = Struct(
    "property_id" / EmergencyLightingTestServerPropertyId,
)

PropertyValue = Switch(
    this.property_id,
    PropertyDict,
)

EmergencyLightingTestServerPropertySet = Struct(
    "property_id" / EmergencyLightingTestServerPropertyId,
    "value" / PropertyValue,
)

# message format is the same
EmergencyLightingTestServerPropertyStatus = EmergencyLightingTestServerPropertySet


EmergencyLightingTestServerParams = SwitchStruct(
    "subopcode" / EnumAdapter(Int8ul, EmergencyLightingTestServerSubOpcode),
    "payload" / Switch(
        this.subopcode,
        {
            EmergencyLightingTestServerSubOpcode.ELT_FUNCTIONAL_TEST_GET: EmergencyLightingTestServerFunctionalTestGet,
            EmergencyLightingTestServerSubOpcode.ELT_FUNCTIONAL_TEST_START: EmergencyLightingTestServerFunctionalTestStart,
            EmergencyLightingTestServerSubOpcode.ELT_FUNCTIONAL_TEST_STOP: EmergencyLightingTestServerFunctionalTestStop,
            EmergencyLightingTestServerSubOpcode.ELT_FUNCTIONAL_TEST_STATUS: EmergencyLightingTestServerFunctionalTestStatus,
            EmergencyLightingTestServerSubOpcode.ELT_DURATION_TEST_GET: EmergencyLightingTestServerDurationTestGet,
            EmergencyLightingTestServerSubOpcode.ELT_DURATION_TEST_START: EmergencyLightingTestServerDurationTestStart,
            EmergencyLightingTestServerSubOpcode.ELT_DURATION_TEST_STOP: EmergencyLightingTestServerDurationTestStop,
            EmergencyLightingTestServerSubOpcode.ELT_DURATION_TEST_STATUS: EmergencyLightingTestServerDurationTestStatus,
            EmergencyLightingTestServerSubOpcode.ELT_PROPERTY_GET: EmergencyLightingTestServerPropertyGet,
            EmergencyLightingTestServerSubOpcode.ELT_PROPERTY_SET: EmergencyLightingTestServerPropertySet,
            EmergencyLightingTestServerSubOpcode.ELT_PROPERTY_SET_UNACKNOWLEDGED: EmergencyLightingTestServerPropertySet,
            EmergencyLightingTestServerSubOpcode.ELT_PROPERTY_STATUS: EmergencyLightingTestServerPropertyStatus,
        }
    )
)

EmergencyLightingTestServerMessage = SwitchStruct(
    "opcode" / Opcode(EmergencyLightingTestServerOpcode),
    "params" / Switch(
        this.opcode,
        {
            EmergencyLightingTestServerOpcode.SILVAIR_ELT: EmergencyLightingTestServerParams
        }
    )
)
# fmt: on



#!python

import importlib
import math
import pkgutil
import sys
from collections import defaultdict
from enum import IntEnum
from functools import partial
from itertools import chain, count

from construct import (
    Array,
    BitsInteger,
    Bytes,
    BytesInteger,
    Computed,
    Construct,
    Flag,
    FocusedSeq,
    FormatField,
    GreedyBytes,
    GreedyRange,
    IfThenElse,
    Pass,
    Renamed,
    Select,
    StopIf,
    StringEncoded,
    Struct,
    Switch,
)

from bluetooth_mesh.messages import AccessMessage


class VariableNames:
    def __init__(self, root_name):
        self.names = {}
        self._load_names(root_name)

    def _load_names(self, root_name):
        root_module = importlib.import_module(root_name)

        for _loader, name, is_pkg in pkgutil.walk_packages(root_module.__path__):
            module_name = f"{root_name}.{name}"
            module = importlib.import_module(module_name)
            self.names.update({id(v): k for k, v in module.__dict__.items()})

            if is_pkg:
                self._load_names(module_name)

    def __getitem__(self, item):
        return self.names.get(id(item))


names = VariableNames("bluetooth_mesh.messages")


def convert(
    con,
    visitor,
    field_name=None,
    struct_name=None,
    many=False,
    message_name=None,
):
    struct_name = struct_name or names[con] or f"{message_name}Params"

    if field_name and field_name.startswith("_"):
        return

    elif isinstance(con, Computed):
        return

    if isinstance(con, Struct):
        visitor.enter_struct(field_name, struct_name, many=many)

        for subcon in con.subcons:
            convert(subcon, visitor, message_name=message_name)

        visitor.exit()

    elif isinstance(con, Switch):
        if many:
            visitor.enter_struct(field_name, struct_name, many=many)
            visitor.enter_union(None)
        else:
            visitor.enter_union(field_name)

        for key, subcon in con.cases.items():
            field_name = type(key)(key).name if isinstance(key, IntEnum) else key
            convert(subcon, visitor, field_name=field_name, message_name=message_name)

        visitor.exit()

        if many:
            visitor.exit()

    elif isinstance(con, Select):
        if many:
            visitor.enter_struct(field_name, struct_name, many=many)
            visitor.enter_union(None)
        else:
            visitor.enter_union(field_name)

        for subcon in con.subcons:
            convert(
                subcon,
                visitor,
                message_name=message_name,
            )

        visitor.exit()

        if many:
            visitor.exit()

    elif isinstance(con, Renamed):
        convert(
            con.subcon,
            visitor,
            field_name=con.name,
            message_name=message_name,
            many=many,
        )

    elif isinstance(con, IfThenElse) and con.elsesubcon is Pass:
        visitor.field(con.thensubcon, field_name, None)

    elif isinstance(con, StringEncoded):
        visitor.field(con, field_name, struct_name)

    elif isinstance(con, (GreedyRange, Array)):
        convert(
            con.subcon,
            visitor,
            field_name=field_name,
            # struct_name=struct_name,
            many=True,
            message_name=message_name,
        )

    elif isinstance(con, FocusedSeq):
        subcon = getattr(con, con.parsebuildfrom)
        convert(
            subcon.subcon,
            visitor,
            field_name=field_name,
            struct_name=struct_name,
            many=False,
        )

    elif con is Pass:
        return

    elif isinstance(con, StopIf):
        return

    elif isinstance(con, Construct):
        subcon = getattr(con, "_subcon", getattr(con, "subcon", None))
        if subcon is None:
            visitor.field(con, field_name, struct_name, many=many)
            return

        convert(
            subcon,
            visitor,
            field_name=field_name,
            struct_name=struct_name,
            message_name=message_name,
            many=many,
        )

    else:
        raise TypeError


class Visitor:
    def __init__(self, struct_name):
        self.types = {}
        self.structs = defaultdict(dict)

        struct_fields = self.structs.setdefault(struct_name, {})
        self.stack = [(struct_name, struct_fields)]
        self._has_union = False

    @property
    def current(self):
        field_name, struct_fields = self.stack[-1]
        return struct_fields

    @staticmethod
    def _camelcase(field_name):
        if field_name is None:
            return None

        head, *tail = str(field_name).lower().replace(" ", "_").split("_")
        return "".join([head, *(i.title() for i in tail)])

    @staticmethod
    def make_type(con):
        FORMAT_FIELD_TYPES = dict(
            c="UInt8",
            b="Int8",
            B="UInt8",
            h="Int16",
            H="UInt16",
            i="Int32",
            I="UInt32",
            l="Int32",
            L="UInt32",
            q="Int64",
            Q="UInt64",
            f="Float32",
            d="Float64",
        )

        if isinstance(con, StringEncoded):
            return "Text"

        elif isinstance(con, Bytes) or con is GreedyBytes:
            return "Data"

        elif isinstance(con, FormatField):
            return FORMAT_FIELD_TYPES[con.fmtstr[1:]]

        elif isinstance(con, BitsInteger):
            width = max(2 ** math.ceil(math.log2(con.length)), 8)
            return f"Int{width}" if con.signed else f"UInt{width}"

        elif isinstance(con, BytesInteger):
            width = max(2 ** math.ceil(math.log2(con.length * 8)), 8)
            return f"Int{width}" if con.signed else f"UInt{width}"

        elif con is Flag:
            return "Bool"

    def enter_struct(self, field_name="", struct_name="", many=False):
        field_name = self._camelcase(field_name)

        self.current[field_name] = f"List({struct_name})" if many else struct_name
        struct_fields = self.structs.setdefault(struct_name, {})
        self.stack.append((struct_name, struct_fields))

    def enter_union(self, field_name=""):
        self._has_union = True
        field_name = self._camelcase(field_name)

        union_fields = self.current.setdefault(field_name, {})
        self.stack.append((field_name, union_fields))

    def exit(self):
        if len(self.stack) == 3:
            if not self._has_union:
                struct_name, struct_fields = self.stack[-1]
                self.types[struct_name] = struct_fields
                self.structs.pop(struct_name)

            self._has_union = False

        self.stack.pop(-1)

    def items(self):
        return chain(self.types.items(), self.structs.items())

    def field(self, con, field_name, struct_name, many=False):
        field_name = self._camelcase(field_name)

        type_name = self.make_type(con) or struct_name

        if many:
            type_name = f"List({type_name})"

        self.current[field_name] = type_name


def generate(protocol_id, file=sys.stdout):
    _print = partial(print, file=file, flush=True)
    visitor = Visitor("AccessMessage")

    for opcode, message in AccessMessage.OPCODES.items():
        convert(message, visitor, field_name=names[opcode], message_name=names[message])

    visitor.structs["AccessMessage"] = {"opcode": "UInt32", None: {}}

    for _opcode, message in AccessMessage.OPCODES.items():
        message_fields = visitor.structs.pop(names[message], visitor.types.get(names[message]))
        if message_fields and None in message_fields and isinstance(message_fields[None], dict):
            visitor.structs["AccessMessage"][None].update(message_fields[None])

    _print(f"@0x{protocol_id:x};")
    _print("")

    def describe(struct_fields, field_tag=None, level=1):
        field_tag = field_tag or count()

        for field_name, field_type in struct_fields.items():
            if isinstance(field_type, dict):
                if field_name:
                    _print(" " * (level * 4 - 1), f"{field_name} :union {{")
                else:
                    _print(" " * (level * 4 - 1), "union {")

                describe(field_type, field_tag, level + 1)

                _print(" " * (level * 4 - 1), "}")
            else:
                _print(
                    " " * (level * 4 - 1),
                    f"{field_name} @{next(field_tag)} :{field_type};",
                    file=file,
                )

    for struct_name, struct_fields in visitor.items():
        _print(f"struct {struct_name} {{")
        describe(struct_fields)
        _print("}")
        _print("")



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from enum import IntEnum

from construct import Embedded, Int8ul, Int16ul, Struct, Switch, this

from bluetooth_mesh.messages.config import StatusCodeAdapter
from bluetooth_mesh.messages.generics import OptionalSetParameters, TransitionTime, TransitionTimeAdapter
from bluetooth_mesh.messages.util import NamedSelect, Opcode, SwitchStruct


class LightLightnessOpcode(IntEnum):
    LIGHT_LIGHTNESS_GET = 0x824B
    LIGHT_LIGHTNESS_SET = 0x824C
    LIGHT_LIGHTNESS_SET_UNACKNOWLEDGED = 0x824D
    LIGHT_LIGHTNESS_STATUS = 0x824E
    LIGHT_LIGHTNESS_LINEAR_GET = 0x824F
    LIGHT_LIGHTNESS_LINEAR_SET = 0x8250
    LIGHT_LIGHTNESS_LINEAR_SET_UNACKNOWLEDGED = 0x8251
    LIGHT_LIGHTNESS_LINEAR_STATUS = 0x8252
    LIGHT_LIGHTNESS_LAST_GET = 0x8253
    LIGHT_LIGHTNESS_LAST_STATUS = 0x8254
    LIGHT_LIGHTNESS_DEFAULT_GET = 0x8255
    LIGHT_LIGHTNESS_DEFAULT_STATUS = 0x8256
    LIGHT_LIGHTNESS_RANGE_GET = 0x8257
    LIGHT_LIGHTNESS_RANGE_STATUS = 0x8258

    def __repr__(self):
        return str(self.value)


class LightLightnessSetupOpcode(IntEnum):
    LIGHT_LIGHTNESS_SETUP_DEFAULT_SET = 0x8259
    LIGHT_LIGHTNESS_SETUP_DEFAULT_SET_UNACKNOWLEDGED = 0x825A
    LIGHT_LIGHTNESS_SETUP_RANGE_SET = 0x825B
    LIGHT_LIGHTNESS_SETUP_RANGE_SET_UNACKNOWLEDGED = 0x825C

    def __repr__(self):
        return str(self.value)


# fmt: off
LightLightnessGet = Struct()

LightLightnessDefault = Struct(
    "lightness" / Int16ul
)

LightLightnessStatusMinimal = Struct(
    "present_lightness" / Int16ul
)

LightLightnessStatusOptional = Struct(
    Embedded(LightLightnessStatusMinimal),
    "target_lightness" / Int16ul,
    "remaining_time" / TransitionTimeAdapter(TransitionTime, allow_unknown=True)
)

LightLightnessStatus = NamedSelect(
    optional=LightLightnessStatusOptional,
    minimal=LightLightnessStatusMinimal
)

LightLightnessRange = Struct(
    "range_min" / Int16ul,
    "range_max" / Int16ul,
)

LightLightnessRangeStatus = Struct(
    "status" / StatusCodeAdapter,
    Embedded(LightLightnessRange)
)

LightLightnessSetMinimal = Struct(
    "lightness" / Int16ul,
    "tid" / Int8ul
)

LightLightnessSetOptional = Struct(
    Embedded(LightLightnessSetMinimal),
    Embedded(OptionalSetParameters)
)

LightLightnessSet = NamedSelect(
    optional=LightLightnessSetOptional,
    minimal=LightLightnessSetMinimal
)

LightLightnessMessage = SwitchStruct(
    "opcode" / Opcode(LightLightnessOpcode),
    "params" / Switch(
        this.opcode,
        {
    LightLightnessOpcode.LIGHT_LIGHTNESS_GET: LightLightnessGet,
    LightLightnessOpcode.LIGHT_LIGHTNESS_SET: LightLightnessSet,
    LightLightnessOpcode.LIGHT_LIGHTNESS_SET_UNACKNOWLEDGED: LightLightnessSet,
    LightLightnessOpcode.LIGHT_LIGHTNESS_STATUS: LightLightnessStatus,
    LightLightnessOpcode.LIGHT_LIGHTNESS_LINEAR_GET: LightLightnessGet,
    LightLightnessOpcode.LIGHT_LIGHTNESS_LINEAR_SET: LightLightnessSet,
    LightLightnessOpcode.LIGHT_LIGHTNESS_LINEAR_SET_UNACKNOWLEDGED: LightLightnessSet,
    LightLightnessOpcode.LIGHT_LIGHTNESS_LINEAR_STATUS: LightLightnessStatus,
    LightLightnessOpcode.LIGHT_LIGHTNESS_LAST_GET: LightLightnessGet,
    LightLightnessOpcode.LIGHT_LIGHTNESS_LAST_STATUS: LightLightnessDefault,
    LightLightnessOpcode.LIGHT_LIGHTNESS_DEFAULT_GET: LightLightnessGet,
    LightLightnessOpcode.LIGHT_LIGHTNESS_DEFAULT_STATUS: LightLightnessDefault,
    LightLightnessOpcode.LIGHT_LIGHTNESS_RANGE_GET: LightLightnessGet,
    LightLightnessOpcode.LIGHT_LIGHTNESS_RANGE_STATUS: LightLightnessRangeStatus,
        }
    )
)

LightLightnessSetupMessage = SwitchStruct(
    "opcode" / Opcode(LightLightnessSetupOpcode),
    "params" / Switch(
        this.opcode,
        {
            LightLightnessSetupOpcode.LIGHT_LIGHTNESS_SETUP_DEFAULT_SET: LightLightnessDefault,
            LightLightnessSetupOpcode.LIGHT_LIGHTNESS_SETUP_DEFAULT_SET_UNACKNOWLEDGED: LightLightnessDefault,
            LightLightnessSetupOpcode.LIGHT_LIGHTNESS_SETUP_RANGE_SET: LightLightnessRange,
            LightLightnessSetupOpcode.LIGHT_LIGHTNESS_SETUP_RANGE_SET_UNACKNOWLEDGED: LightLightnessRange,
        }
    )
)
# fmt: on



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import importlib
import json
import logging
from tempfile import NamedTemporaryFile

import pytest

from bluetooth_mesh.messages import AccessMessage
from bluetooth_mesh.messages.util import to_camelcase_dict, to_snakecase_dict

valid = [
    # debug
    bytes.fromhex("f5360100"),  # RSSI_THRESHOLD_GET
    bytes.fromhex("f5360101f0"),  # RSSI_THRESHOLD_SET
    bytes.fromhex("f5360102f0"),  # RSSI_THRESHOLD_STATUS
    bytes.fromhex("f5360103ff"),  # RADIO_TEST
    bytes.fromhex("f536010a"),  # UPTIME_GET
    bytes.fromhex("f536010b00000000"),  # UPTIME_STATUS
    bytes.fromhex("f536010bffffffff"),  # UPTIME_STATUS
    bytes.fromhex("f536010c"),  # LAST_SW_FAULT_GET
    bytes.fromhex("f536010d"),  # LAST_SW_FAULT_CLEAR
    bytes.fromhex("f536010e00000000616263"),  # LAST_SW_FAULT_STATUS
    bytes.fromhex("f536010f"),  # SYSTEM_STATS_GET
    bytes.fromhex("f5360110"),  # SYSTEM_STATS_STATUS
    bytes.fromhex("f53601106162636465666768000000000000"),
    bytes.fromhex("f53601106162636400000000ffff00000000"),
    bytes.fromhex("f5360111"),  # LAST_MALLOC_FAULT_GET
    bytes.fromhex("f5360112"),  # LAST_MALLOC_FAULT_CLEAR
    bytes.fromhex("f536011300000000616263"),  # LAST_MALLOC_FAULT_STATUS
    bytes.fromhex("f5360114"),  # LAST_FDS_FAULT_GET
    bytes.fromhex("f5360115"),  # LAST_FDS_FAULT_CLEAR
    bytes.fromhex("f536011600000000616263"),  # LAST_FDS_FAULT_STATUS
    bytes.fromhex("f5360117"),  # BYTES_BEFORE_GARBAGE_COLLECTOR_GET
    bytes.fromhex("f53601180000"),  # BYTES_BEFORE_GARBAGE_COLLECTOR_STATUS
    bytes.fromhex("f53601181234"),  # BYTES_BEFORE_GARBAGE_COLLECTOR_STATUS
    bytes.fromhex("f5360118ffff"),  # BYTES_BEFORE_GARBAGE_COLLECTOR_STATUS
    bytes.fromhex("f5360119"),  # PROVISIONED_APP_VERSION_GET
    bytes.fromhex("f536011a0000"),  # PROVISIONED_APP_VERSION_STATUS
    bytes.fromhex("f536011a1234"),  # PROVISIONED_APP_VERSION_STATUS
    bytes.fromhex("f536011affff"),  # PROVISIONED_APP_VERSION_STATUS
    bytes.fromhex("f536011b"),  # FULL_FIRMWARE_VERSION_GET
    bytes.fromhex("f536011c61626364"),  # FULL_FIRMWARE_VERSION_STATUS
    bytes.fromhex("f536011d"),  # IV_INDEX_GET
    bytes.fromhex("f536011e00000000"),  # IV_INDEX_STATUS
    bytes.fromhex("f536011e12345678"),  # IV_INDEX_STATUS
    bytes.fromhex("f536011effffffff"),  # IV_INDEX_STATUS
    bytes.fromhex("f536011f"),  # GARBAGE_COLLECTOR_COUNTER_GET
    bytes.fromhex("f53601200000"),  # GARBAGE_COLLECTOR_COUNTER_STATUS
    bytes.fromhex("f53601201234"),  # GARBAGE_COLLECTOR_COUNTER_STATUS
    bytes.fromhex("f5360120ffff"),  # GARBAGE_COLLECTOR_COUNTER_STATUS
    bytes.fromhex("f5360121"),  # ARAP_LIST_SIZE_GET
    bytes.fromhex("f53601220000"),  # ARAP_LIST_SIZE_STATUS
    bytes.fromhex("f53601221234"),  # ARAP_LIST_SIZE_STATUS
    bytes.fromhex("f5360122ffff"),  # ARAP_LIST_SIZE_STATUS
    bytes.fromhex("f536012200000000"),  # ARAP_LIST_SIZE_STATUS
    bytes.fromhex("f536012212345678"),  # ARAP_LIST_SIZE_STATUS
    bytes.fromhex("f5360122ffffffff"),  # ARAP_LIST_SIZE_STATUS
    bytes.fromhex("f536012300"),  # ARAP_LIST_CONTENT_GET
    bytes.fromhex("f536012312"),  # ARAP_LIST_CONTENT_GET
    bytes.fromhex("f5360123ff"),  # ARAP_LIST_CONTENT_GET
    bytes.fromhex("f53601240000"),  # ARAP_LIST_CONTENT_STATUS
    bytes.fromhex("f536012400000000000000"),  # ARAP_LIST_CONTENT_STATUS
    bytes.fromhex("f53601240000ffffffffff"),  # ARAP_LIST_CONTENT_STATUS
    bytes.fromhex("f53601240000ffffffffff00000000001234561234"),  # ARAP_LIST_CONTENT_STATUS
    # config
    bytes.fromhex("02003601CE00FECAEFBE0BB000000000"),
    bytes.fromhex("8002000b00010000012100"),
    bytes.fromhex("8039010203040506070809"),
    bytes.fromhex("803E00A70202000213"),
    # ctl
    bytes.fromhex("826522223333ff323c"),
    # nds
    bytes.fromhex("fd360100"),  # PUBLICATION_GET
    bytes.fromhex("fd3601011234aaaa82031000"),  # PUBLICATION_SET
    bytes.fromhex("fd3601011234aaaa820310000300"),  # PUBLICATION_SET
    bytes.fromhex("fd3601011234aaaa82031000"),  # PUBLICATION_SET
    bytes.fromhex("fd3601021234aaaa820310000300"),  # PUBLICATION_STATUS
    bytes.fromhex("fc360100"),  # SUBSCRIPTION_GET
    bytes.fromhex("fc3601010123aaaa"),  # SUBSCRIPTION_SET
    bytes.fromhex("fc3601020123aaaa"),  # SUBSCRIPTION_SET_UNACKNOWLEDGED
    bytes.fromhex("fc3601031234aaaa20012300110101"),  # SUBSCRIPTION_STATUS
    # sensor
    bytes.fromhex("8230"),
    bytes.fromhex("82300400"),
    bytes.fromhex("8231"),
    bytes.fromhex("82310700"),
    bytes.fromhex("510c00000000040b0c"),
    bytes.fromhex("511900"),
    bytes.fromhex("510c00000000020b0c1f00efcdab071b1c"),
    bytes.fromhex("52e20ac800"),
    bytes.fromhex("52220b2003"),
    bytes.fromhex("52440da244ff"),
    ##vendor sensor
    # bytes.fromhex("52099040a244ff0000"),
    bytes.fromhex("52440da244ff220b2003"),
    bytes.fromhex("583000010004000900"),
    bytes.fromhex("583000"),
    bytes.fromhex("5957005700c800"),
    bytes.fromhex("5b5700570001c800"),
    bytes.fromhex("5b5700020001c80039"),
    bytes.fromhex("5957005700c800"),
    bytes.fromhex("5957000200c80039"),
    bytes.fromhex("5905000500200354"),
    bytes.fromhex("5957000200c80000"),
    bytes.fromhex("59590059000003"),
    bytes.fromhex("5942004d0001"),
    bytes.fromhex("594200420050"),
    bytes.fromhex("590a003600b80b00"),
    bytes.fromhex("596d006d000a0000"),
    ##no value
    # bytes.fromhex("596d006d00ffffff"),
    bytes.fromhex("59550055001a2700"),
    bytes.fromhex("594c004c001b1a"),
    bytes.fromhex("596c006c00ff1b1a"),
    ##no value
    # bytes.fromhex("596c006c00ffffff"),
    bytes.fromhex("59680068000500"),
    bytes.fromhex("5967006700050001007040"),
    bytes.fromhex("590e000e006162636465666768"),
    bytes.fromhex("590e000e006162636465660000"),
    bytes.fromhex("5911001100616263646566676861626364656667686162636465666768616263646566676861626364"),
    bytes.fromhex("596a006a00a244ff"),
    bytes.fromhex("592e002e0044ff"),
    bytes.fromhex("593200320044ff0000"),
    bytes.fromhex("5952005200a08601"),
    bytes.fromhex("5916001600e80300d007000f2700"),
    bytes.fromhex("594f004f001f"),
    bytes.fromhex("594f004f00e1"),
    bytes.fromhex("5954005400e620"),
    bytes.fromhex("59450045003102dc6e71"),
    bytes.fromhex("591400140092096400f8f8204e71"),
    bytes.fromhex("5913001300dc6e"),
    bytes.fromhex("5901000100186da2"),
    bytes.fromhex("5960006000a244ff6da2"),
    bytes.fromhex("592a002a00690140000000f0ff54"),
    bytes.fromhex("594900490000000006f0ff"),
    bytes.fromhex("59470047009709640000007d1571"),
    bytes.fromhex("594600460000007d15"),
    bytes.fromhex("5921002100000001007d15"),
    bytes.fromhex("5970007000020000"),
    bytes.fromhex("59060006000020"),
    bytes.fromhex("5940004000020000"),
    bytes.fromhex("591f001f00d007"),
    bytes.fromhex("5941004100e803d007"),
    bytes.fromhex("593e003e00d407"),
    bytes.fromhex("590f000f00ffeeddccbbaa"),
    bytes.fromhex("590700070064"),
    bytes.fromhex("595e005e009227"),
    bytes.fromhex("5951005100b80b"),
    bytes.fromhex("590a000a0004f0"),
    bytes.fromhex("590b000b002a00"),
    bytes.fromhex("590c000c00de4600"),
    bytes.fromhex("5950005000ee00cdab"),
    bytes.fromhex("590800080064"),
    bytes.fromhex("59080008009c"),
    bytes.fromhex("596100610088aa00bbbb"),
    bytes.fromhex("5962006200881a27001a2700"),
    bytes.fromhex("596400640088ffffe620"),
    bytes.fromhex("5965006500880000cdab"),
    bytes.fromhex("59660066008820032003"),
    # light
    bytes.fromhex("824b"),
    bytes.fromhex("824cbbaa22"),
    bytes.fromhex("824c010022"),
    bytes.fromhex("824c000031323c"),
    bytes.fromhex("824d000031323c"),
    bytes.fromhex("824e4400"),
    bytes.fromhex("824e000031c80f"),
    bytes.fromhex("824f"),
    bytes.fromhex("8250bbaa01"),
    bytes.fromhex("8250010022"),
    bytes.fromhex("8250000031321b"),
    bytes.fromhex("8251ff0031323c"),
    bytes.fromhex("82520000ddbb4c"),
    bytes.fromhex("8253"),
    bytes.fromhex("82540000"),
    bytes.fromhex("8255"),
    bytes.fromhex("82560000"),
    bytes.fromhex("8257"),
    bytes.fromhex("82580011118888"),
    # scene
    bytes.fromhex("8241"),
    bytes.fromhex("824201001e"),
    bytes.fromhex("824301001e"),
    bytes.fromhex("824201001ef23c"),
    bytes.fromhex("824301001ef23c"),
    bytes.fromhex("5e000100"),
    bytes.fromhex("5e0001000200f2"),
    bytes.fromhex("8244"),
    bytes.fromhex("824500010001000200") + 14 * bytes.fromhex("0000"),
    bytes.fromhex("82460100"),
    bytes.fromhex("82470100"),
    bytes.fromhex("829e0100"),
    bytes.fromhex("829f0100"),
    # onoff
    bytes.fromhex("8201"),
    bytes.fromhex("82020122"),
    bytes.fromhex("82020122"),
    bytes.fromhex("82020031323c"),
    bytes.fromhex("82020031f23c"),
    bytes.fromhex("820400"),
    bytes.fromhex("820400014a"),
    bytes.fromhex("82040001ff"),
    # level
    bytes.fromhex("8205"),
    bytes.fromhex("8206ff7f22"),
    bytes.fromhex("8206008022"),
    bytes.fromhex("8206010022"),
    bytes.fromhex("8206000031323c"),
    bytes.fromhex("8207000031323c"),
    bytes.fromhex("8208ff7f"),
    bytes.fromhex("82080080"),
    bytes.fromhex("82080000ff004a"),
    bytes.fromhex("820800000100ff"),
    bytes.fromhex("8209ffffff7f22"),
    bytes.fromhex("82090000008022"),
    bytes.fromhex("82090100000022"),
    bytes.fromhex("82090000000031323c"),
    bytes.fromhex("820a0000000031323c"),
    bytes.fromhex("820bff7f22"),
    bytes.fromhex("820b008022"),
    bytes.fromhex("820b010022"),
    bytes.fromhex("820b000031323c"),
    bytes.fromhex("820c000031323c"),
    # Time Status
    bytes.fromhex("5d8ea9282a005905490248"),
    # debugV2
    bytes.fromhex("EB360102090002001020300208000420"),
    bytes.fromhex("EB3601020b020c06001000200030"),
    bytes.fromhex("EB3601024b000b096d7920737472696e67"),
    bytes.fromhex("EB360102550002001020300a000800320028"),
    bytes.fromhex("EB3601031300"),
    bytes.fromhex("EB3601000f00"),
    bytes.fromhex("EB360102fe00ff"),
    # config - Composition Data Page 1 - Status
    bytes.fromhex("0201010000"),
    bytes.fromhex("020101000400"),
    bytes.fromhex("020102010005000000010201000501170101"),
    bytes.fromhex("0201050307011AFD0901383001080508480000050A00050AD0"),
    # config - Relay - Status
    bytes.fromhex("80280200"),
    bytes.fromhex("802802FF"),
    bytes.fromhex("80280100"),
    bytes.fromhex("80280000"),
    bytes.fromhex("802801FF"),
    bytes.fromhex("802800FF"),
    # config - Relay - Set
    bytes.fromhex("80270100"),
    bytes.fromhex("80270000"),
    bytes.fromhex("802701FF"),
    bytes.fromhex("802700FF"),
    # Silvair ELT Server
    bytes.fromhex("E936010305040302010608070506"),
]


@pytest.fixture(scope="session")
def capnproto():
    import capnp

    from bluetooth_mesh.messages.capnproto import generate

    with NamedTemporaryFile("w", suffix=".capnp") as f:
        generate(0xD988DA1AAFBE9E47, f)
        f.flush()
        return capnp.load(f.name)


@pytest.mark.skipif(not importlib.util.find_spec("capnp"), reason="requires Python3.7")
@pytest.mark.parametrize("encoded", [pytest.param(i, id=i.hex()) for i in valid])
def test_parse_capnproto(encoded, capnproto):
    logging.info("MESH[%i] %s", len(encoded), encoded.hex())

    decoded = AccessMessage.parse(encoded)
    logging.info("CONSTRUCT %r", decoded)

    params = to_camelcase_dict(decoded)
    logging.info("CAPNP INPUT[%i] %s", len(json.dumps(params)), json.dumps(params))

    message = capnproto.AccessMessage.new_message(**params)
    logging.info("CAPNP %r", message)

    packed = message.to_bytes_packed()
    logging.info("PACKED[%i] %s", len(packed), packed.hex())

    unpacked = capnproto.AccessMessage.from_bytes_packed(packed)
    logging.info("UNPACKED %r", unpacked)

    params = to_snakecase_dict(unpacked.to_dict())
    logging.info("CONSTRUCT INPUT %s", params)

    recoded = AccessMessage.build(params)
    logging.info("RECODED[%i] %s", len(recoded), recoded.hex())

    assert recoded == encoded



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from enum import IntEnum

from construct import (
    BitsInteger,
    BitStruct,
    ByteSwapped,
    Bytewise,
    GreedyRange,
    GreedyString,
    Int8ul,
    Int16ul,
    Int24ub,
    Int32ul,
    PaddedString,
    Padding,
    Struct,
    this,
)

from bluetooth_mesh.messages.util import EnumAdapter, EnumSwitch as Switch, NamedSelect, Opcode, SwitchStruct


class DebugOpcode(IntEnum):
    SILVAIR_DEBUG = 0xF53601


class DebugSubOpcode(IntEnum):
    RSSI_THRESHOLD_GET = 0x00
    RSSI_THRESHOLD_SET = 0x01
    RSSI_THRESHOLD_STATUS = 0x02
    RADIO_TEST = 0x03
    UPTIME_GET = 0x0A
    UPTIME_STATUS = 0x0B
    LAST_SW_FAULT_GET = 0x0C
    LAST_SW_FAULT_CLEAR = 0x0D
    LAST_SW_FAULT_STATUS = 0x0E
    SYSTEM_STATS_GET = 0x0F
    SYSTEM_STATS_STATUS = 0x10
    LAST_MALLOC_FAULT_GET = 0x11
    LAST_MALLOC_FAULT_CLEAR = 0x12
    LAST_MALLOC_FAULT_STATUS = 0x13
    LAST_FDS_FAULT_GET = 0x14
    LAST_FDS_FAULT_CLEAR = 0x15
    LAST_FDS_FAULT_STATUS = 0x16
    BYTES_BEFORE_GARBAGE_COLLECTOR_GET = 0x17
    BYTES_BEFORE_GARBAGE_COLLECTOR_STATUS = 0x18
    PROVISIONED_APP_VERSION_GET = 0x19
    PROVISIONED_APP_VERSION_STATUS = 0x1A
    FULL_FIRMWARE_VERSION_GET = 0x1B
    FULL_FIRMWARE_VERSION_STATUS = 0x1C
    IV_INDEX_GET = 0x1D
    IV_INDEX_STATUS = 0x1E
    GARBAGE_COLLECTOR_COUNTER_GET = 0x1F
    GARBAGE_COLLECTOR_COUNTER_STATUS = 0x20
    ARAP_LIST_SIZE_GET = 0x21
    ARAP_LIST_SIZE_STATUS = 0x22
    ARAP_LIST_CONTENT_GET = 0x23
    ARAP_LIST_CONTENT_STATUS = 0x24


# fmt: off
RssiThresholdGet = Struct()

RssiThreshold = Struct(
    "rssi_threshold" / Int8ul
)

RadioTest = Struct(
    "packet_counter" / Int8ul
)

UptimeGet = Struct()

UptimeStatus = Struct(
    "uptime" / Int32ul
)

LastSwFaultGet = Struct()

LastSwFaultClear = Struct()

LastSwFaultStatus = Struct(
    "time" / Int32ul,
    "fault" / GreedyString("utf8")
)

SystemStatsGet = Struct()

SystemStat = Struct(
    "name" / PaddedString(8, "utf8"),
    "high_water_mark" / Int32ul,
    "_rfu" / Padding(2)
)

SystemStatsStatus = Struct(
    "stats" / GreedyRange(SystemStat)
)

LastMallocFaultGet = Struct()

LastMallocFaultClear = Struct()

LastMallocFaultStatus = LastSwFaultStatus

LastFdsFaultGet = Struct()

LastFdsFaultClear = Struct()

LastFdsFaultStatus = LastSwFaultStatus

BytesBeforeGarbageCollectorGet = Struct()

BytesBeforeGarbageCollectorStatus = Struct(
    "bytes_left" / Int16ul
)

ProvisionedAppVersionGet = Struct()

ProvisionedAppVersionStatus = Struct(
    "version" / Int16ul
)

FullFirmwareVersionGet = Struct()

FullFirmwareVersionStatus = Struct(
    "version" / GreedyString("utf8")
)

IvIndexGet = Struct()

IvIndexStatus = Struct(
    "ivindex" / Int32ul
)

GarbageCollectorCounterGet = Struct()

GarbageCollectorCounterStatus = Struct(
    "counter" / Int16ul
)

ArapSize16 = Struct(
    "capacity" / Int16ul,
    "size" / Int16ul
)

ArapSize8 = Struct(
    "capacity" / Int8ul,
    "size" / Int8ul
)

ArapListSizeGet = Struct()

ArapListSizeStatus = NamedSelect(
    new=ArapSize16,
    old=ArapSize8,
)

ArapListContentGet = Struct(
    "page" / Int8ul
)

ArapNode = ByteSwapped(
    BitStruct(
        "sequence" / Bytewise(Int24ub),
        "ivi" / BitsInteger(1),
        "address" / BitsInteger(15),
    )
)

ArapListContentStatus = Struct(
    "current_page" / Int8ul,
    "last_page" / Int8ul,
    "nodes" / GreedyRange(ArapNode)
)

DebugParams = SwitchStruct(
    "subopcode" / EnumAdapter(Int8ul, DebugSubOpcode),
    "payload" / Switch(
        this.subopcode,
        {
            DebugSubOpcode.RSSI_THRESHOLD_GET: RssiThresholdGet,
            DebugSubOpcode.RSSI_THRESHOLD_SET: RssiThreshold,
            DebugSubOpcode.RSSI_THRESHOLD_STATUS: RssiThreshold,
            DebugSubOpcode.RADIO_TEST: RadioTest,
            DebugSubOpcode.UPTIME_GET: UptimeGet,
            DebugSubOpcode.UPTIME_STATUS: UptimeStatus,
            DebugSubOpcode.LAST_SW_FAULT_GET: LastSwFaultGet,
            DebugSubOpcode.LAST_SW_FAULT_CLEAR: LastSwFaultClear,
            DebugSubOpcode.LAST_SW_FAULT_STATUS: LastSwFaultStatus,
            DebugSubOpcode.SYSTEM_STATS_GET: SystemStatsGet,
            DebugSubOpcode.SYSTEM_STATS_STATUS: SystemStatsStatus,
            DebugSubOpcode.LAST_MALLOC_FAULT_GET: LastMallocFaultGet,
            DebugSubOpcode.LAST_MALLOC_FAULT_CLEAR: LastMallocFaultClear,
            DebugSubOpcode.LAST_MALLOC_FAULT_STATUS: LastMallocFaultStatus,
            DebugSubOpcode.LAST_FDS_FAULT_GET: LastFdsFaultGet,
            DebugSubOpcode.LAST_FDS_FAULT_CLEAR: LastFdsFaultClear,
            DebugSubOpcode.LAST_FDS_FAULT_STATUS: LastFdsFaultStatus,
            DebugSubOpcode.BYTES_BEFORE_GARBAGE_COLLECTOR_GET: BytesBeforeGarbageCollectorGet,
            DebugSubOpcode.BYTES_BEFORE_GARBAGE_COLLECTOR_STATUS: BytesBeforeGarbageCollectorStatus,
            DebugSubOpcode.PROVISIONED_APP_VERSION_GET: ProvisionedAppVersionGet,
            DebugSubOpcode.PROVISIONED_APP_VERSION_STATUS: ProvisionedAppVersionStatus,
            DebugSubOpcode.FULL_FIRMWARE_VERSION_GET: FullFirmwareVersionGet,
            DebugSubOpcode.FULL_FIRMWARE_VERSION_STATUS: FullFirmwareVersionStatus,
            DebugSubOpcode.IV_INDEX_GET: IvIndexGet,
            DebugSubOpcode.IV_INDEX_STATUS: IvIndexStatus,
            DebugSubOpcode.GARBAGE_COLLECTOR_COUNTER_GET: GarbageCollectorCounterGet,
            DebugSubOpcode.GARBAGE_COLLECTOR_COUNTER_STATUS: GarbageCollectorCounterStatus,
            DebugSubOpcode.ARAP_LIST_SIZE_GET: ArapListSizeGet,
            DebugSubOpcode.ARAP_LIST_SIZE_STATUS: ArapListSizeStatus,
            DebugSubOpcode.ARAP_LIST_CONTENT_GET: ArapListContentGet,
            DebugSubOpcode.ARAP_LIST_CONTENT_STATUS: ArapListContentStatus,
        }
    )
)

DebugMessage = SwitchStruct(
    "opcode" / Opcode(DebugOpcode),
    "params" / Switch(
        this.opcode,
        {
            DebugOpcode.SILVAIR_DEBUG: DebugParams,
        }
    )
)
# fmt: on



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from copy import deepcopy

import pytest
from construct import ValidationError

from bluetooth_mesh.messages.config import *

valid = [
    # fmt: off
    pytest.param(
        SecureNetworkBeaconAdapter,
        bytes.fromhex('00'),
        SecureNetworkBeacon.OFF,
        id="SecureNetworkBeaconAdapter - Off"
    ),
    pytest.param(
        SecureNetworkBeaconAdapter,
        bytes.fromhex('01'),
        SecureNetworkBeacon.ON,
        id="SecureNetworkBeaconAdapter - On"
    ),
    pytest.param(
        GATTProxyAdapter,
        bytes.fromhex('00'),
        GATTProxy.DISABLED,
        id="GATTProxyAdapter - Disabled"
    ),
    pytest.param(
        GATTProxyAdapter,
        bytes.fromhex('01'),
        GATTProxy.ENABLED,
        id="GATTProxyAdapter - Enabled"
    ),
    pytest.param(
        GATTProxyAdapter,
        bytes.fromhex('02'),
        GATTProxy.NOT_SUPPORTED,
        id="GATTProxyAdapter - NOT_SUPPORTED"
    ),
    pytest.param(
        RelayAdapter,
        bytes.fromhex('00'),
        Relay.DISABLED,
        id="RelayAdapter - Disabled"
    ),
    pytest.param(
        RelayAdapter,
        bytes.fromhex('01'),
        Relay.ENABLED,
        id="RelayAdapter - Enabled"
    ),
    pytest.param(
        RelayAdapter,
        bytes.fromhex('02'),
        Relay.NOT_SUPPORTED,
        id="RelayAdapter - NOT_SUPPORTED"
    ),
    pytest.param(
        GATTNamespaceDescriptorAdapter,
        bytes.fromhex('DC00'),
        GATTNamespaceDescriptor.TWO_HUNDRED_TWENTIETH,
        id="GATTNamespaceDescriptorAdapter - TWO_HUNDRED_TWENTIETH"
    ),
    pytest.param(
        GATTNamespaceDescriptorAdapter,
        bytes.fromhex('0000'),
        GATTNamespaceDescriptor.UNKNOWN,
        id="GATTNamespaceDescriptorAdapter - UNKNOWN"
    ),
    pytest.param(
        TTL,
        bytes.fromhex('7F'),
        0x7F,
        id="TTL-Max"
    ),
    pytest.param(
        TTL,
        bytes.fromhex('6F'),
        0x6F,
        id="TTL"
    ),
    pytest.param(
        TTL,
        bytes.fromhex('00'),
        0x00,
        id="TTL-Min"
    ),
    pytest.param(
        SIGModelId,
        bytes.fromhex('ADDE'),
        {
            "model_id": 0xDEAD,
        },
        id="SIGModelId"
    ),
    pytest.param(
        SIGModelId,
        bytes.fromhex('ADDE'),
        {
            "model_id": 0xDEAD,
            "vendor_id": 0xC0DE
        },
        id="SIGModelId - accepts None vendor id"
    ),
    pytest.param(
        VendorModelId,
        bytes.fromhex('DEC0ADDE'),
        {
            "model_id": 0xDEAD,
            "vendor_id": 0xC0DE
        },
        id="VendorModelId"
    ),
    pytest.param(
        ModelId,
        bytes.fromhex('DEC0ADDE'),
        {
            "model_id": 0xDEAD,
            "vendor_id": 0xC0DE
        },
        id="ModelId - vendor"
    ),
    pytest.param(
        ModelId,
        bytes.fromhex('ADDE'),
        {
            "model_id": 0xDEAD
        },
        id="ModelId - SIG"
    ),
    pytest.param(
        ModelId,
        bytes.fromhex('ADDE'),
        {
            "model_id": 0xDEAD,
            "vendor_id": None
        },
        id="ModelId - accepts None vendor id"
    ),
    pytest.param(
        UnassignedAddress,
        bytes.fromhex('0000'),
        0x0000,
        id="UnassignedAddress"
    ),
    pytest.param(
        UnicastAddress,
        bytes.fromhex('FF7F'),
        0x7FFF,
        id="UnicastAddress - Max"
    ),
    pytest.param(
        UnicastAddress,
        bytes.fromhex('0100'),
        0x0001,
        id="UnicastAddress - Min"
    ),
    pytest.param(
        GroupAddress,
        bytes.fromhex('00C0'),
        0xC000,
        id="GroupAddress - Min"
    ),
    pytest.param(
        VirtualLabel,
        bytes.fromhex('0080'),
        0x8000,
        id="VirtualAddress - Min"
    ),
    pytest.param(
        Retransmit,
        bytes.fromhex('07'),
        {
            "count": 0x07,
            "interval_steps": 0x00
        },
        id="Retransmit - Max count"
    ),
    pytest.param(
        Retransmit,
        bytes.fromhex('F8'),
        {
            "count": 0x00,
            "interval_steps": 0x1F
        },
        id="Retransmit - Max steps"
    ),
    pytest.param(
        NetworkRetransmit,
        bytes.fromhex('07'),
        {
            "count": 0x07,
            "interval": 10
        },
        id="NetworkRetransmit - Max count"
    ),
    pytest.param(
        NetworkRetransmit,
        bytes.fromhex('F8'),
        {
            "count": 0x00,
            "interval": 320
        },
        id="NetworkRetransmit - Max steps"
    ),
    pytest.param(
        RelayRetransmit,
        bytes.fromhex('07'),
        {
            "count": 0x07,
            "interval": 10
        },
        id="RelayRetransmit - Max count"
    ),
    pytest.param(
        RelayRetransmit,
        bytes.fromhex('F8'),
        {
            "count": 0x00,
            "interval": 320
        },
        id="RelayRetransmit - Max steps"
    ),
    pytest.param(
        PublishRetransmit,
        bytes.fromhex('07'),
        {
            "count": 0x07,
            "interval": 50
        },
        id="PublishRetransmit - Max count"
    ),
    pytest.param(
        PublishRetransmit,
        bytes.fromhex('F8'),
        {
            "count": 0x00,
            "interval": 1600
        },
        id="PublishRetransmit - Max steps"
    ),
    pytest.param(
        Struct(*NetAndAppKeyIndex),
        bytes.fromhex('012345'),
        {
            "net_key_index": 0x301,
            "app_key_index": 0x452,
        },
        id="NetAndAppKeyIndex"
    ),
    pytest.param(
        Struct(*NetAndAppKeyIndex),
        bytes.fromhex('abcdef'),
        {
            "net_key_index": 0xdab,
            "app_key_index": 0xefc
        },
        id="NetAndAppKeyIndex"
    ),
    pytest.param(
        Struct(*NetAndAppKeyIndex),
        bytes.fromhex('efcdab'),
        {
            "net_key_index": 0xdef,
            "app_key_index": 0xabc
        },
        id="NetAndAppKeyIndex"
    ),
    pytest.param(
        PublishPeriod,
        bytes.fromhex('00'),
        {
            "step_resolution": PublishPeriodStepResolution.RESOLUTION_100_MS,
            "number_of_steps": 0x00
        },
        id="PublishPeriod"
    ),
    pytest.param(
        PublishPeriod,
        bytes.fromhex('C0'),
        {
            "step_resolution": PublishPeriodStepResolution.RESOLUTION_10_MIN,
            "number_of_steps": 0x00
        },
        id="PublishPeriod - Max resolution"
    ),
    pytest.param(
        PublishPeriod,
        bytes.fromhex('3F'),
        {
            "step_resolution": PublishPeriodStepResolution.RESOLUTION_100_MS,
            "number_of_steps": 0x3F
        },
        id="PublishPeriod - Max steps"
    ),
    pytest.param(
        ConfigBeaconGet,
        bytes(),
        {},
        id="ConfigBeaconGet"
    ),
    pytest.param(
        ConfigBeaconSet,
        bytes.fromhex('00'),
        {
            "beacon": SecureNetworkBeacon.OFF
        },
        id="ConfigBeaconSet"
    ),
    pytest.param(
        ConfigBeaconStatus,
        bytes.fromhex('01'),
        {
            "beacon": SecureNetworkBeacon.ON
        },
        id="ConfigBeaconStatus-On"
    ),
    pytest.param(
        ConfigCompositionDataGet,
        bytes.fromhex('00'),
        {
            "page": 0x00
        },
        id="ConfigCompositionDataGet - Min page"
    ),
    pytest.param(
        ConfigCompositionDataGet,
        bytes.fromhex('FF'),
        {
            "page": 0xFF
        },
        id="ConfigCompositionDataGet - Max page"
    ),
    pytest.param(
        ConfigCompositionDataStatus,
        bytes.fromhex('003601CE00FECAEFBE0BB000000000'),
        {
            "page": 0x00,
            "data": {
                "cid": 0x0136,
                "pid": 0x00CE,
                "vid": 0xCAFE,
                "crpl": 0xBEEF,
                "features": 0xB00B,
                "elements": [
                    {
                        "location": GATTNamespaceDescriptor.UNKNOWN,
                        "sig_number": 0x00,
                        "vendor_number": 0x00,
                        "sig_models": [],
                        "vendor_models": []
                    }
                ]
            },
            "zero": {
                "cid": 0x0136,
                "pid": 0x00CE,
                "vid": 0xCAFE,
                "crpl": 0xBEEF,
                "features": 0xB00B,
                "elements": [
                    {
                        "location": GATTNamespaceDescriptor.UNKNOWN,
                        "sig_number": 0x00,
                        "vendor_number": 0x00,
                        "sig_models": [],
                        "vendor_models": []
                    }
                ]
            }
        },
        id="ConfigCompositionDataStatus - page 0", ),
    pytest.param(
        ConfigCompositionDataStatus,
        bytes.fromhex('010100010A'),
        {
            "page": 0x01,
            "data":
                {"element": [
                    {
                        "number_s": 1,
                        "number_v": 0,
                        "sig_models": [
                            {
                                "extended_items_count": 0,
                                "format": ExtendedModelsItemFormat.SHORT,
                                "corresponding_present": True,
                                "corresponding_id": 0x0A,
                                "extended_models_items": {"short": []}
                            },
                        ],
                        "vendor_models": []
                    },
                ],
                },

            "first":
                {"element": [
                    {
                        "number_s": 1,
                        "number_v": 0,
                        "sig_models": [
                            {
                                "extended_items_count": 0,
                                "format": ExtendedModelsItemFormat.SHORT,
                                "corresponding_present": True,
                                "corresponding_id": 0x0A,
                                "extended_models_items": {"short": []}
                            }
                        ],
                        "vendor_models": []
                    }
                ]
                }
        },
        id="ConfigCompositionDataStatus - page 1"
    ),
    pytest.param(
        ConfigCompositionDataStatus,
        bytes.fromhex('02001601000001010000'),
        {
            "page": 0x02,
            "second":
                {"record_list": [
                    {
                        "mesh_profile_identifier": 0x1600,
                        "version": {
                            "version_x": 1,
                            "version_y": 0,
                            "version_z": 0
                        },
                        "num_element_offsets": 1,
                        "element_offset_list": [1],
                        "additional_data_len": 0,
                        "additional_data": []
                    }
                ]}
        },
        id="ConfigCompositionDataStatus - page 2"
    ),
    pytest.param(
        ConfigCompositionDataStatus,
        bytes.fromhex('FFCAFE'),
        {
            "page": 0xFF,
            "two_hundred_and_fifty_fifth": bytes.fromhex('CAFE'),
        },
        id="ConfigCompositionDataStatus - not page 0 or 1 or 3"
    ),
    pytest.param(
        CompositionDataPage0,
        bytes.fromhex('3601CE00FECAEFBE0BB000000000'),
        {
            "cid": 0x0136,
            "pid": 0x00CE,
            "vid": 0xCAFE,
            "crpl": 0xBEEF,
            "features": 0xB00B,
            "elements": [
                {
                    "location": GATTNamespaceDescriptor.UNKNOWN,
                    "sig_number": 0x00,
                    "vendor_number": 0x00,
                    "sig_models": [],
                    "vendor_models": []
                }
            ]
        },
        id="CompositionData - One element", ),
    pytest.param(
        CompositionDataPage0,
        bytes.fromhex('3601CE00FECAEFBE0BB00000000000000000'),
        {
            "cid": 0x0136,
            "pid": 0x00CE,
            "vid": 0xCAFE,
            "crpl": 0xBEEF,
            "features": 0xB00B,
            "elements": [
                {
                    "location": GATTNamespaceDescriptor.UNKNOWN,
                    "sig_number": 0x00,
                    "vendor_number": 0x00,
                    "sig_models": [],
                    "vendor_models": []
                },
                {
                    "location": GATTNamespaceDescriptor.UNKNOWN,
                    "sig_number": 0x00,
                    "vendor_number": 0x00,
                    "sig_models": [],
                    "vendor_models": []
                }
            ]
        },
        id="CompositionData - Two elements"
    ),
    pytest.param(
        CompositionDataPage0Element,
        bytes.fromhex('00000101ADDEEFBEADDE'),
        {
            "location": GATTNamespaceDescriptor.UNKNOWN,
            "sig_number": 0x01,
            "vendor_number": 0x01,
            "sig_models": [{"model_id": 0xDEAD}],
            "vendor_models": [{"model_id": 0xDEAD, "vendor_id": 0xBEEF}]
        },
        id="CompositionDataPage0Element", ),
    pytest.param(
        CompositionDataPage0Element,
        bytes.fromhex('00000001EFBEADDE'),
        {
            "location": GATTNamespaceDescriptor.UNKNOWN,
            "sig_number": 0x00,
            "vendor_number": 0x01,
            "sig_models": [],
            "vendor_models": [{"model_id": 0xDEAD, "vendor_id": 0xBEEF}]
        },
        id="CompositionDataPage0Element - No SIG", ),
    pytest.param(
        CompositionDataPage0Element,
        bytes.fromhex('00000100ADDE'),
        {
            "location": GATTNamespaceDescriptor.UNKNOWN,
            "sig_number": 0x01,
            "vendor_number": 0x00,
            "sig_models": [{"model_id": 0xDEAD}],
            "vendor_models": []
        },
        id="CompositionDataPage0Element - No vendor"
    ),

    pytest.param(
        CompositionDataPage1,
        bytes.fromhex('02010005000000010201000501170101'),
        {"element": [
            {
                "number_s": 2,
                "number_v": 1,
                "sig_models": [
                    {
                        "extended_items_count": 0,
                        "format": ExtendedModelsItemFormat.SHORT,
                        "corresponding_present": False,
                        "corresponding_id": 0,
                        "extended_models_items": {"short": []}
                    },
                    {
                        "extended_items_count": 1,
                        "format": ExtendedModelsItemFormat.SHORT,
                        "corresponding_present": True,
                        "corresponding_id": 0,
                        "extended_models_items": {
                            "short": [
                                {
                                    "model_item_index": 0,
                                    "element_offset": 0
                                }
                            ]
                        }
                    },
                ],
                "vendor_models": [
                    {
                        "extended_items_count": 0,
                        "format": ExtendedModelsItemFormat.SHORT,
                        "corresponding_present": False,
                        "corresponding_id": 0,
                        "extended_models_items": {"short": []}
                    },
                ]
            },
            {
                "number_s": 1,
                "number_v": 2,
                "sig_models": [
                    {
                        "extended_items_count": 0,
                        "format": ExtendedModelsItemFormat.SHORT,
                        "corresponding_present": True,
                        "corresponding_id": 0,
                        "extended_models_items": {"short": []}
                    },
                ],
                "vendor_models": [
                    {
                        "extended_items_count": 1,
                        "format": ExtendedModelsItemFormat.SHORT,
                        "corresponding_present": True,
                        "corresponding_id": 1,
                        "extended_models_items": {
                            "short": [
                                {
                                    "model_item_index": 2,
                                    "element_offset": -1
                                }
                            ]
                        }
                    },
                    {
                        "extended_items_count": 0,
                        "format": ExtendedModelsItemFormat.SHORT,
                        "corresponding_present": True,
                        "corresponding_id": 1,
                        "extended_models_items": {"short": []}
                    }

                ]
            }
        ]
        },
        id="CompositionData Page 1"
    ),
    pytest.param(
        CompositionDataPage1Element,
        bytes.fromhex('050307011AFD0901383001080508480000050A00050AD0'),
        {
            "number_s": 5,
            "number_v": 3,
            "sig_models": [
                {
                    "extended_items_count": 1,
                    "format": ExtendedModelsItemFormat.LONG,
                    "corresponding_present": True,
                    "corresponding_id": 0x01,
                    "extended_models_items": {
                        "long": [
                            {
                                "model_item_index": 26,
                                "element_offset": -3
                            }
                        ]
                    }
                },
                {
                    "extended_items_count": 2,
                    "format": ExtendedModelsItemFormat.SHORT,
                    "corresponding_present": True,
                    "corresponding_id": 0x01,
                    "extended_models_items": {
                        "short": [
                            {
                                "model_item_index": 0x07,
                                "element_offset": 0x00
                            },
                            {
                                "model_item_index": 0x06,
                                "element_offset": 0x00
                            }
                        ]
                    }
                },
                {
                    "extended_items_count": 0,
                    "format": ExtendedModelsItemFormat.SHORT,
                    "corresponding_present": True,
                    "corresponding_id": 0x08,
                    "extended_models_items": {"short": []}
                },
                {
                    "extended_items_count": 1,
                    "format": ExtendedModelsItemFormat.SHORT,
                    "corresponding_present": True,
                    "corresponding_id": 0x08,
                    "extended_models_items": {
                        "short": [
                            {
                                "model_item_index": 0x09,
                                "element_offset": 0x00
                            }
                        ]
                    }
                },
                {
                    "extended_items_count": 0,
                    "format": ExtendedModelsItemFormat.SHORT,
                    "corresponding_present": False,
                    "corresponding_id": 0,
                    "extended_models_items": {"short": []}
                },
            ],
            "vendor_models": [
                        {
                            "extended_items_count": 0,
                            "format": ExtendedModelsItemFormat.SHORT,
                            "corresponding_present": False,
                            "corresponding_id": 0,
                            "extended_models_items": {"short": []}
                        },
                        {
                            "extended_items_count": 1,
                            "format": ExtendedModelsItemFormat.SHORT,
                            "corresponding_present": True,
                            "corresponding_id": 0x0A,
                            "extended_models_items": {
                                "short": [
                                    {
                                        "model_item_index": 0x00,
                                        "element_offset": 0x00
                                    }
                                ]
                            }
                        },
                        {
                            "extended_items_count": 1,
                            "format": ExtendedModelsItemFormat.SHORT,
                            "corresponding_present": True,
                            "corresponding_id": 0x0A,
                            "extended_models_items": {
                                "short": [
                                    {
                                        "model_item_index": 0x1A,
                                        "element_offset": 0x00
                                    }
                                ]
                            }
                        },
                    ]
        },
        id="Config Composition Data Page 1 - single element"
    ),
    pytest.param(
        CompositionDataPage2,
        bytes.fromhex('021601020303010203020011FF'),
        {"record_list": [
            {
                "mesh_profile_identifier": 0x1602,
                "version": {
                    "version_x": 1,
                    "version_y": 2,
                    "version_z": 3
                },
                "num_element_offsets": 3,
                "element_offset_list": [1, 2, 3],
                "additional_data_len": 2,
                "additional_data":  [0x11, 0xFF]

            }]
        },
        id="CompositionData Page 2"
    ),
    pytest.param(
        ModelRelationItem,
        bytes.fromhex('010A'),
        {
            "extended_items_count": 0,
            "format": ExtendedModelsItemFormat.SHORT,
            "corresponding_present": True,
            "corresponding_id": 10,
            "extended_models_items": {"short": []}
        },
        id="Config Composition Data Page 1 - Single model item in short format with corresponding"
    ),
    pytest.param(
        ModelRelationItem,
        bytes.fromhex('030B'),
        {
            "extended_items_count": 0,
            "format": ExtendedModelsItemFormat.LONG,
            "corresponding_present": True,
            "corresponding_id": 11,
            "extended_models_items": {"long": []}
        },
        id="Config Composition Data Page 1 - Single model item in long format with corresponding"
    ),
    pytest.param(
        ModelRelationItem,
        bytes.fromhex('0458'),
        {
            "extended_items_count": 1,
            "format": ExtendedModelsItemFormat.SHORT,
            "corresponding_present": False,
            "corresponding_id": 0,
            "extended_models_items": {
                "short": [
                    {
                        "model_item_index": 0x0B,
                        "element_offset": 0x00
                    }
                ]
            }
        },
        id="Config Composition Data Page 1 - Single model item in short format with extending"
    ),
    pytest.param(
        ModelRelationItem,
        bytes.fromhex('060A0B'),
        {
            "extended_items_count": 1,
            "format": ExtendedModelsItemFormat.LONG,
            "corresponding_present": False,
            "corresponding_id": 0,
            "extended_models_items": {
                "long": [
                    {
                        "model_item_index": 10,
                        "element_offset": 11
                    }
                ]
            }
        },
        id="Config Composition Data Page 1 - Single model item in long format with extending"
    ),
    pytest.param(
        ConfigDefaultTTLGet,
        bytes(),
        {},
        id="ConfigDefaultTTLGet"
    ),
    pytest.param(
        ConfigDefaultTTLSet,
        bytes.fromhex('00'),
        {
            "ttl": 0x00
        },
        id="ConfigDefaultTTLSet - Min"
    ),
    pytest.param(
        ConfigDefaultTTLSet,
        bytes.fromhex('0B'),
        {
            "ttl": 0x0B
        },
        id="ConfigDefaultTTLSet"
    ),
    pytest.param(
        ConfigDefaultTTLSet,
        bytes.fromhex('7F'),
        {
            "ttl": 0x7F
        },
        id="ConfigDefaultTTLSet - Max"
    ),
    pytest.param(
        ConfigDefaultTTLStatus,
        bytes.fromhex('00'),
        {
            "ttl": 0x00
        },
        id="ConfigDefaultTTLStatus - Min"
    ),
    pytest.param(
        ConfigDefaultTTLStatus,
        bytes.fromhex('0B'),
        {
            "ttl": 0x0B
        },
        id="ConfigDefaultTTLStatus"
    ),
    pytest.param(
        ConfigDefaultTTLStatus,
        bytes.fromhex('7F'),
        {
            "ttl": 0x7F
        },
        id="ConfigDefaultTTLStatus - Max"
    ),
    pytest.param(
        ConfigGATTProxyGet,
        bytes(),
        {},
        id="ConfigGATTProxyGet"
    ),
    pytest.param(
        ConfigGATTProxySet,
        bytes.fromhex('00'),
        {
            "GATT_proxy": GATTProxy.DISABLED
        },
        id="ConfigGATTProxySet Disabled"
    ),
    pytest.param(
        ConfigGATTProxyStatus,
        bytes.fromhex('01'),
        {
            "GATT_proxy": GATTProxy.ENABLED
        },
        id="ConfigGATTProxyStatus Enabled"
    ),
    pytest.param(
        ConfigRelayGet,
        bytes(),
        {},
        id="ConfigRelayGet"
    ),
    pytest.param(
        ConfigRelaySet,
        bytes.fromhex('0000'),
        {
            "relay": Relay.DISABLED,
            "retransmit": {
                "count": 0x00,
                "interval": 10
            }
        },
        id="ConfigRelaySet"
    ),
    pytest.param(
        ConfigRelayStatus,
        bytes.fromhex('0200'),
        {
            "relay": Relay.NOT_SUPPORTED,
            "retransmit": {
                "count": 0x00,
                "interval": 10
            }
        },
        id="ConfigRelayStatus"
    ),
    pytest.param(
        ConfigModelPublicationGet,
        bytes.fromhex('0102 0304'),
        {
            "element_address": 0x0201,
            "model": {
                "model_id": 0x0403
            }
        },
        id="ConfigModelPublicationGet - SIG"
    ),
    pytest.param(
        ConfigModelPublicationGet,
        bytes.fromhex('010203040506'),
        {
            "element_address": 0x0201,
            "model": {
                "model_id": 0x0605,
                "vendor_id": 0x0403
            }
        },
        id="ConfigModelPublicationGet - Vendor"
    ),
    pytest.param(
        ConfigModelPublicationSet,
        bytes.fromhex('0102 0100 BC1A 7FC00703040506'),
        {
            "element_address": 0x0201,
            "publish_address": 0x0001,
            "rfu": 0,
            "credential_flag": PublishFriendshipCredentialsFlag.FRIENDSHIP_SECURITY,
            "app_key_index": 0xabc,
            "ttl": 0x7F,
            "publish_period": {
                "step_resolution": PublishPeriodStepResolution.RESOLUTION_10_MIN,
                "number_of_steps": 0x00
            },
            "retransmit": {
                "count": 0x07,
                "interval": 50
            },
            "model": {
                "model_id": 0x0605,
                "vendor_id": 0x0403
            }
        },
        id="ConfigModelPublicationSet"
    ),
    pytest.param(
        ConfigModelPublicationSet,
        bytes.fromhex('0201 0403 0500 0607110403'),
        {
            "element_address": 0x0102,
            "publish_address": 0x0304,
            "rfu": 0,
            "credential_flag": PublishFriendshipCredentialsFlag.MASTER_SECURITY,
            "app_key_index": 5,
            "ttl": 6,
            "publish_period": {
                "step_resolution": PublishPeriodStepResolution.RESOLUTION_100_MS,
                "number_of_steps": 7
            },
            "retransmit": {
                "count": 1,
                "interval": 150
            },
            "model": {
                "model_id": 0x0304
            }
        },
        id="ConfigModelPublicationSet"
    ),
    pytest.param(
        ConfigModelPublicationStatus,
        bytes.fromhex('02 0201 0403 0500 06 07 11 0403 0605'),
        {
            "status": StatusCode.INVALID_MODEL,
            "element_address": 0x0102,
            "publish_address": 0x0304,
            "rfu": 0,
            "credential_flag": PublishFriendshipCredentialsFlag.MASTER_SECURITY,
            "app_key_index": 5,
            "ttl": 6,
            "publish_period": {
                "step_resolution": PublishPeriodStepResolution.RESOLUTION_100_MS,
                "number_of_steps": 7
            },
            "retransmit": {
                "count": 1,
                "interval": 150
            },
            "model": {
                "model_id": 0x0506,
                "vendor_id": 0x0304
            }
        },
        id="ConfigModelPublicationStatus"
    ),
    pytest.param(
        ConfigModelSubscriptionAdd,
        bytes.fromhex('AA11 BBC2 CC33'),
        {
            "element_address": 0x11AA,
            "address": 0xC2BB,
            "model": {
                "model_id": 0x33CC
            }
        },
        id="ConfigModelSubscriptionAddSigId"
    ),
    pytest.param(
        ConfigModelSubscriptionAdd,
        bytes.fromhex('0201 FDFF 3F00 2A00'),
        {
            "element_address": 0x0102,
            "address": 0xFFFD,
            "model": {
                "model_id": 0x002A,
                "vendor_id": 0x003F
            }
        },
        id="ConfigModelSubscriptionAddVendorId"
    ),
    pytest.param(
        ConfigModelSubscriptionStatus,
        bytes.fromhex('00 AA11 0000 CC33'),
        {
            "status": StatusCode.SUCCESS,
            "element_address": 0x11AA,
            "address": 0,
            "model": {
                "model_id": 0x33CC
            }
        },
        id="ConfigModelSubscriptionStatus[UNASSIGNED]"
    ),
    pytest.param(
        ConfigModelSubscriptionStatus,
        bytes.fromhex('00 AA11 0080 CC33'),
        {
            "status": StatusCode.SUCCESS,
            "element_address": 0x11AA,
            "address": 0x8000,
            "model": {
                "model_id": 0x33CC
            }
        },
        id="ConfigModelSubscriptionStatus[VIRTUAL]"
    ),
    pytest.param(
        Struct(*SingleKeyIndex('key_index')),
        bytes.fromhex('bc0a'),
        {
            "key_index": 0xabc,
        },
        id="SingleKeyIndex"
    ),
    pytest.param(
        ConfigAppKeyAdd,
        bytes.fromhex('012345 000102030405060708090A0B0C0D0E0F'),
        {
            "net_key_index": 0x301,
            "app_key_index": 0x452,
            "app_key": bytes.fromhex('000102030405060708090A0B0C0D0E0F'),
        },
        id="ConfigAppKeyAdd"
    ),
    pytest.param(
        ConfigAppKeyAdd,
        bytes.fromhex('236145 63964771734fbd76e3b40519d1d94a48'),
        {
            "net_key_index": 0x123,
            "app_key_index": 0x456,
            "app_key": bytes.fromhex('63964771734fbd76e3b40519d1d94a48'),
        },
        id="ConfigAppKeyAdd"
    ),
    pytest.param(
        ConfigAppKeyGet,
        bytes.fromhex('0200'),
        {
            "net_key_index": 2,
        },
        id="ConfigAppKeyGet"
    ),
    pytest.param(
        ConfigAppKeyStatus,
        bytes.fromhex('00 332322'),
        {
            "status": StatusCode.SUCCESS,
            "net_key_index": 0x333,
            "app_key_index": 0x222
        },
        id="ConfigAppKeyStatus"
    ),
    pytest.param(
        ConfigAppKeyList,
        bytes.fromhex('000b00 010000 012100'),
        {
            "status": StatusCode.SUCCESS,
            "net_key_index": 11,
            "app_key_indices": [0, 1, 2, 257]
        },
        id="ConfigAppKeyList_even"
    ),
    pytest.param(
        ConfigAppKeyList,
        bytes.fromhex('000b00 563412 8907'),
        {
            "status": StatusCode.SUCCESS,
            "net_key_index": 11,
            "app_key_indices": [0x123, 0x456, 0x789]
        },
        id="ConfigAppKeyList_odd"
    ),
    pytest.param(
        ConfigNetKeyList,
        bytes.fromhex('42b000 '),
        {
            "net_key_indices": [11, 66],
        },
        id="ConfigNetKeyList_even"
    ),
    pytest.param(
        ConfigNetKeyAdd,
        bytes.fromhex('4305 000102030405060708090A0B0C0D0E0F'),
        {
            "net_key_index": 0x543,
            "net_key": bytes.fromhex('000102030405060708090A0B0C0D0E0F'),
        },
        id="ConfigNetKeyAdd"
    ),
    pytest.param(
        ConfigNetKeyList,
        bytes.fromhex('43d002 5800'),
        {
            "net_key_indices": [45, 67, 88],

        },
        id="ConfigNetKeyList_odd"
    ),
    pytest.param(
        ConfigNodeIdentitySet,
        bytes.fromhex('FF0F01'),
        {
            "net_key_index": 0xFFF,
            "identity": NodeIdentity.RUNNING
        },
        id="ConfigNodeIdentitySet"
    ),
    pytest.param(
        ConfigNodeIdentityStatus,
        bytes.fromhex('00FF0F01'),
        {
            "status": StatusCode.SUCCESS,
            "net_key_index": 0xFFF,
            "identity": NodeIdentity.RUNNING
        },
        id="ConfigNodeIdentityStatus"
    ),
    pytest.param(
        ConfigHeartbeatPublicationGet,
        bytes(),
        {},
        id="ConfigHeartbeatPublicationGet"
    ),
    pytest.param(
        ConfigHeartbeatPublicationSet,
        bytes.fromhex('010203040506070809'),
        {
            "destination": 0x0201,
            "count": 4,
            "period": 8,
            "ttl": 0x05,
            "features": {5, 6, 13, 14, 15},
            "net_key_index": 0x908,
        },
        id="ConfigHeartbeatPublicationSet"
    ),
    pytest.param(
        ConfigHeartbeatPublicationSet,
        bytes.fromhex('0102ff060506070809'),
        {
            "destination": 0x0201,
            "count": float('inf'),
            "period": 32,
            "ttl": 0x05,
            "features": {5, 6, 13, 14, 15},
            "net_key_index": 0x908,
        },
        id="ConfigHeartbeatPublicationSet - infinite count"
    ),
    pytest.param(
        ConfigHeartbeatPublicationSet,
        bytes.fromhex('0102ff100506070809'),
        {
            "destination": 0x0201,
            "count": float('inf'),
            "period": 0x8000,
            "ttl": 0x05,
            "features": {5, 6, 13, 14, 15},
            "net_key_index": 0x908,
        },
        id="ConfigHeartbeatPublicationSet - long period"
    ),
    pytest.param(
        ConfigNetworkTransmitSet,
        bytes.fromhex('f9'),
        {
            "count": 0x01,
            "interval": 320
        },
        id="ConfigNetworkTransmitSet"
    ),
    # fmt: on
]

build_valid = [
    # fmt: off
    pytest.param(
        CompositionDataPage0Element,
        bytes.fromhex('00000101ADDEEFBEADDE'),
        {
            "location": GATTNamespaceDescriptor.UNKNOWN,
            "sig_models": [{
                "model_id": 0xDEAD
            }],
            "vendor_models": [{
                "model_id": 0xDEAD,
                "vendor_id": 0xBEEF
            }]
        },
        id="CompositionDataPage0Element (Vendor and SIG number not needed)"
    ),
    pytest.param(
        CompositionDataPage0Element,
        bytes.fromhex('00000001EFBEADDE'),
        {
            "location": GATTNamespaceDescriptor.UNKNOWN,
            "sig_models": [],
            "vendor_models": [{
                "model_id": 0xDEAD,
                "vendor_id": 0xBEEF
            }]
        },
        id="CompositionDataPage0Element (Vendor and SIG number not needed) - No SIG"
    ),
    pytest.param(
        CompositionDataPage0Element,
        bytes.fromhex('00000100ADDE'),
        {
            "location": GATTNamespaceDescriptor.UNKNOWN,
            "sig_models": [{
                "model_id": 0xDEAD
            }],
            "vendor_models": []
        },
        id="CompositionDataPage0Element (Vendor and SIG number not needed) - No vendor"
    ),
    # fmt: on
]

parse_invalid = [
    # fmt: off
    pytest.param(
        ConfigHeartbeatPublicationSet,
        bytes.fromhex('010203048006070809'),
        ValidationError,
        id="ConfigHeartbeatPublicationSet - TTL too big",
    ),
    pytest.param(
        ConfigHeartbeatPublicationSet,
        bytes.fromhex('010211040506070809'),
        ValidationError,
        id="ConfigHeartbeatPublicationSet - count too big",
    ),
    pytest.param(
        ConfigHeartbeatPublicationSet,
        bytes.fromhex('010203110506070809'),
        ValidationError,
        id="ConfigHeartbeatPublicationSet - period too long",
    ),
    # fmt: on
]

build_invalid = [
    # fmt: off
    pytest.param(
        ConfigHeartbeatPublicationSet,
        {
            "destination": 0x0201,
            "count": 1,
            "period": 0x8001,
            "ttl": 0x05,
            "features": {5, 6, 13, 14, 15},
            "net_key_index": 0x0908
        },
        ValidationError,
        id="ConfigHeartbeatPublicationSet - period too long",
    ),
    pytest.param(
        ConfigHeartbeatPublicationSet,
        {
            "destination": 0x0201,
            "count": 1,
            "period": float('inf'),
            "ttl": 0x05,
            "features": {5, 6, 13, 14, 15},
            "net_key_index": 0x0908
        },
        ValidationError,
        id="ConfigHeartbeatPublicationSet - infinite period",
    ),
    # fmt: on
]


@pytest.mark.parametrize("message,encoded,decoded", valid + build_valid)
def test_build(message, encoded, decoded):
    _decoded = deepcopy(decoded)
    result = message.build(obj=_decoded)
    assert result.hex() == encoded.hex()


@pytest.mark.parametrize("message,encoded,decoded", valid)
def test_parse(message, encoded, decoded):
    result = message.parse(data=encoded)
    assert result == decoded


@pytest.mark.parametrize("message,encoded,exception", parse_invalid)
def test_parse_invalid(message, encoded, exception):
    with pytest.raises(exception):
        message.parse(data=encoded)


@pytest.mark.parametrize("message,decoded,exception", build_invalid)
def test_build_invalid(message, decoded, exception):
    with pytest.raises(exception):
        message.build(obj=decoded)


@pytest.mark.parametrize("key", ["params", "config_appkey_add"])
def test_build_config_message(key):
    app_key = bytes.fromhex("deadbeef" * 4)

    data = ConfigMessage.build(
        {
            "opcode": ConfigOpcode.CONFIG_APPKEY_ADD,
            key: dict(
                app_key_index=1,
                net_key_index=1,
                app_key=app_key,
            ),
        }
    )

    assert data == bytes.fromhex("00011000") + app_key


def test_parse_config_message():
    key = bytes.fromhex("deadbeef" * 4)

    msg = ConfigMessage.parse(bytes.fromhex("00011000") + key)

    assert msg == dict(
        opcode=ConfigOpcode.CONFIG_APPKEY_ADD,
        params=dict(
            app_key_index=1,
            net_key_index=1,
            app_key=key,
        ),
        config_appkey_add=dict(
            app_key_index=1,
            net_key_index=1,
            app_key=key,
        ),
    )


invalid_retr = [
    # fmt: off
    pytest.param(
        NetworkRetransmit,
        bytes.fromhex('F8'),
        {
            "count": 0x00,
            "interval": 0
        },
        id="interval = 0"
    ),
    pytest.param(
        NetworkRetransmit,
        bytes.fromhex('F8'),
        {
            "count": 0x00,
            "interval": 5
        },
        id="interval = 5"
    ),
    pytest.param(
        NetworkRetransmit,
        bytes.fromhex('F8'),
        {
            "count": 0x00,
            "interval": 330
        },
        id="interval too high"
    ),
    pytest.param(
        PublishRetransmit,
        bytes.fromhex('F8'),
        {
            "count": 0x00,
            "interval": 1700
        },
        id="interval too high"
    ),
    pytest.param(
        PublishRetransmit,
        bytes.fromhex('F8'),
        {
            "count": 0x08,
            "interval": 700
        },
        id="count too high"
    ),
]


@pytest.mark.parametrize("message,encoded,decoded", invalid_retr)
def test_invalid_retransmit(message, encoded, decoded):
    _decoded = deepcopy(decoded)
    with pytest.raises(AttributeError):
        message.build(obj=_decoded)



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from enum import IntEnum

from construct import (
    Float32l,
    Float64l,
    GreedyRange,
    Int8sl,
    Int8ul,
    Int16sl,
    Int16ul,
    Int32sl,
    Int32ul,
    Int64sl,
    Int64ul,
    PascalString,
    PrefixedArray,
    Struct,
    this,
)

from bluetooth_mesh.messages.util import EnumAdapter, EnumSwitch as Switch, Opcode, SwitchStruct


class DebugV2Opcode(IntEnum):
    SILVAIR_DEBUG_V2 = 0xEB3601


class DebugV2SubOpcode(IntEnum):
    GET = 0x00
    SET = 0x01
    STATUS = 0x02
    CLEAR = 0x03


class DebugV2ParameterTypes(IntEnum):
    GETABLE_PARAMETER_LIST = 0x00
    SETABLE_PARAMETER_LIST = 0x01
    SETTABLE_PARAMETER_TYPE_LIST = 0x02
    CLEARABLE_PARAMETER_LIST = 0x03
    RSSI_THRESHOLD = 0x08
    UPTIME = 0x09
    MESH_RPL_LIST_CONTENT = 0x0A
    TIME_SYNC_MSG_LIST = 0x0B
    FIRST_TIME_SYNC_TIMESTAMP = 0x0C
    LAST_TIME_SYNC_TIME = 0x0D
    LAST_TIME_RECOVERY_TIME = 0x0E
    RECALLED_EVENTS_TIME = 0x0F
    RESET_MONITOR_COUNTERS = 0x10
    TIME_RECOVERY_COUNTER = 0x11
    OCCUPANCY_EVENTS_COUNTER = 0x12
    OCCUPANCY_HISTORY = 0x13
    OCCUPANCY_HISTORY_RESOLUTION = 0x14
    RADIO_STATS_PER_CHANNEL = 0x15
    QUEUES_MONITOR = 0x16
    QUEUES_STATS = 0x17
    UART_ERRORS = 0x18
    UART_MODEL_INSTANCES = 0x19
    APP_THREAD_LATENCY = 0x1A
    IV_INDEX_INFO = 0x1B
    MESH_CACHE_INFO = 0x1C
    EL_ELT_HISTORY = 0x1D
    RRULE_SCHEDULE_REGISTER_SIZE = 0x1E
    RRULE_NEXT_EVENT_TAI = 0x1F
    RRULE_NEXT_EVENT_ENTRIES_ARRAY = 0x20
    LAST_CALLS_OF_RRULE_ENTRIES = 0x21
    CPU_USAGE_CURRENT_STATS = 0x22
    CPU_USAGE_HISTOGRAM = 0x23
    CPU_USAGE_HISTORY = 0x24
    CPU_USAGE_HISTORY_SETUP = 0x25
    LAST_MEASURED_POWER_OFF_TIME_MS = 0x27
    LAST_MEASURED_DALI_BUS_SHORT_TO_POWER_OFF_TIME_MS = 0x28
    LAST_RESET_FIRST_FAULT_INFO = 0x29
    LC_OLC_AMBIENT_LUX_LEVEL = 0x2A

    NOT_KNOWN_PARAMETER = 0xFF

    @classmethod
    def _missing_(cls, value):
        return DebugV2ParameterTypes.NOT_KNOWN_PARAMETER


class DebugV2DataTypes(IntEnum):
    UINT8 = 0x00
    UINT16 = 0x01
    UINT32 = 0x02
    UINT64 = 0x03
    INT8 = 0x04
    INT16 = 0x05
    INT32 = 0x06
    INT64 = 0x07
    FLOAT = 0x08
    DOUBLE = 0x09
    ENUM = 0x0A
    STRING = 0x0B
    ARRAY = 0x0C
    INVALID = 0xFF


# fmt: off
DebugV2Parameter = SwitchStruct(
    "data_type" / EnumAdapter(Int8ul, DebugV2DataTypes),
    "parameter" / Switch(
        this.data_type,
        {
            DebugV2DataTypes.UINT8: Int8ul,
            DebugV2DataTypes.UINT16: Int16ul,
            DebugV2DataTypes.UINT32: Int32ul,
            DebugV2DataTypes.UINT64: Int64ul,
            DebugV2DataTypes.INT8: Int8sl,
            DebugV2DataTypes.INT16: Int16sl,
            DebugV2DataTypes.INT32: Int32sl,
            DebugV2DataTypes.INT64: Int64sl,
            DebugV2DataTypes.FLOAT: Float32l,
            DebugV2DataTypes.DOUBLE: Float64l,
            DebugV2DataTypes.ENUM: Int8ul,
            DebugV2DataTypes.STRING: PascalString(Int8ul, "utf-8"),
            DebugV2DataTypes.ARRAY: PrefixedArray(Int8ul, Int8ul),
            DebugV2DataTypes.INVALID: Struct()
        }
    )
)

DebugV2SubMessage = Struct(
    "subopcode" / EnumAdapter(Int8ul, DebugV2SubOpcode),
    "parameter_type" / Int8ul,
    "page_number" / Int8ul,
    "payload" / GreedyRange(DebugV2Parameter)
)

DebugV2Message = SwitchStruct(
    "opcode" / Opcode(DebugV2Opcode),
    "params" / Switch(
        this.opcode,
        {
            DebugV2Opcode.SILVAIR_DEBUG_V2: DebugV2SubMessage,
        }
    )
)

# fmt: on



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# pylint: disable=W0223

import enum
import math
import re
import sys
from datetime import date, datetime, timedelta
from ipaddress import IPv4Address

from construct import (
    Adapter,
    Bit,
    BitStruct,
    Bitwise,
    Computed,
    Construct,
    Container,
    Embedded,
    Enum,
    ExprValidator,
    Float64b,
    FuncPath,
    IfThenElse,
    Int32ub,
    Pass,
    Rebuild,
    Restreamed,
    Select,
    SizeofError,
    Struct,
    Switch,
    ValidationError,
    stream_read,
    stream_write,
    this,
)


def identity(x):
    return x


def reverse(sequence):
    return sequence[::-1]


def Reversed(subcon):
    return Restreamed(subcon, reverse, subcon.sizeof(), reverse, subcon.sizeof(), identity)


class BitListAdapter(Adapter):
    def __init__(self, subcon, reversed):
        super().__init__(subcon)
        self.reversed = reversed

    def _decode(self, obj, content, path):
        bits = set()

        for bit, feature in enumerate(reversed(obj) if self.reversed else obj):
            if not feature:
                continue

            bits.add(bit)

        return bits

    def _encode(self, obj, content, path):
        bits = []

        for bit in range(self.sizeof() * 8):
            bits.append(bit in obj)

        return list(reversed(bits)) if self.reversed else bits


def BitList(size, *, reversed=False):
    return BitListAdapter(Bitwise(Bit[size * 8]), reversed)


class SetAdapter(Adapter):
    def _decode(self, obj, context, path):
        return set(obj)

    def _encode(self, obj, context, path):
        return obj


def EnumAdapter(subcon, enum):
    class _Enum(Enum):
        ENUM = enum

    class _EnumAdapter(Adapter):
        type = enum
        _enum = enum

        def _decode(self, obj, context, path):
            if obj not in enum._value2member_map_:
                raise ValidationError("object failed validation: '%s' not in %s" % (obj, enum))
            return enum(obj)

        def _encode(self, obj, context, path):
            if obj == enum:
                return obj.value

            try:
                return enum[obj] if isinstance(obj, str) else enum(obj)
            except ValueError as ex:
                raise ValidationError("object failed validation: '%s' not in %s" % (obj, enum)) from ex

    _EnumAdapter.__construct_doc__ = _Enum(subcon, enum)

    return _EnumAdapter(subcon)


def LogAdapter(subcon, *, max_value=None, infinity=False):
    class _LogAdapter(Adapter):
        MAX_TYPE_VALUE = int(math.pow(2, subcon.length * 8) - 1)
        _subcon = subcon

        def _decode(self, obj, context, path):
            if obj == 0:
                return 0

            if obj == self.MAX_TYPE_VALUE:
                if infinity:
                    return float("inf")

            if max_value is not None and obj > max_value:
                raise ValidationError("max value exceeded, expecting at most %d: %s" % (max_value, obj))

            return int(math.pow(2, obj - 1))

        def _encode(self, obj, context, path):
            if obj == 0:
                return 0

            if obj == float("inf"):
                if infinity:
                    return self.MAX_TYPE_VALUE

                raise ValidationError("infinity is not allowed: %s" % obj)

            value = math.log(obj, 2) + 1

            if max_value is not None and value > max_value:
                raise ValidationError("max value exceeded, expecting at most %d: %s" % (max_value, obj))

            return int(value)

    return _LogAdapter(subcon)


def RangeValidator(subcon, *, min_value=None, max_value=None):
    def validate_range(obj, ctx):
        if min_value is not None and obj < min_value:
            return False

        if max_value is not None and obj > max_value:
            return False

        return True

    return ExprValidator(subcon, validate_range)


class FieldAdapter(Adapter):
    def __init__(self, subcon, field):
        self._subcon = field
        super().__init__(subcon)

    def _decode(self, obj, content, path):
        return obj

    def _encode(self, obj, content, path):
        return obj


def EmbeddedBitStruct(name, *fields, reversed=False):
    """
    Emulates BitStruct embedding:
        - for parsing, adds Computed accessor fields to the parent construct,
        - for building, Rebuild the bit struct using keys passed to the parent

    NOTE: This is a hack. Do not use unless you absolutely have to.
    """
    bit_struct = BitStruct(*fields)

    if reversed:
        bit_struct = Reversed(bit_struct)

    bit_struct.__construct_doc__ = Embedded(Struct(*fields))

    return (
        name / Rebuild(bit_struct, dict),
        *(
            field.name / FieldAdapter(Computed(this[name][field.name]), field)
            for field in fields
            if field.name is not None
        ),
    )


class Opcode(Construct):
    __construct_doc__ = Int32ub
    subcon = Int32ub

    def __init__(self, type=int):
        super().__init__()
        self.type = type

    def _parse(self, stream, context, path):
        try:
            opcode = stream_read(stream, 1)[0]

            if opcode == 0x7F:
                raise ValidationError

            len = opcode >> 7

            # 1 byte opcode
            if not len:
                return self.type(opcode)

            len = opcode >> 6
            opcode = opcode << 8 | stream_read(stream, 1)[0]

            # 2 byte opcode
            if len == 2:
                return self.type(opcode)

            if len == 3:
                opcode = opcode << 8 | stream_read(stream, 1)[0]
                return self.type(opcode)

            raise ValidationError
        except ValueError as ex:
            raise ValidationError from ex

    def _build(self, obj, stream, context, path):
        if obj > 0xFFFF:
            stream_write(stream, obj.to_bytes(3, byteorder="big"))
        elif obj > 0xFF:
            stream_write(stream, obj.to_bytes(2, byteorder="big"))
        else:
            stream_write(stream, obj.to_bytes(1, byteorder="big"))

        return self.type(obj)

    def _sizeof(self, context, path):
        raise SizeofError


class DefaultCountValidator(Adapter):
    _subcon = Float64b

    def __init__(self, subcon, rounding=None, resolution=1.0, unknown_value=True):
        super().__init__(subcon)
        self.rounding = rounding
        self.resolution = resolution
        self.unknown_value = unknown_value

    def _decode(self, obj, content, path):
        if self.unknown_value and obj == (256**self.subcon.length) - 1:
            return float(sys.float_info.max)
        else:
            return round(obj * self.resolution, self.rounding) if self.rounding else obj * self.resolution

    def _encode(self, obj, content, path):
        if self.unknown_value and obj == float(sys.float_info.max):
            return (256**self.subcon.length) - 1
        else:
            return round(obj / self.resolution)


class MacAddressAdapter(Adapter):
    def _decode(self, obj, context, path):
        return ":".join(map("{:02x}".format, obj))

    def _encode(self, obj, context, path):
        return bytes(int(i, 16) for i in obj.split(":"))


class IpAddressAdapter(Adapter):
    def _decode(self, obj, context, path):
        return IPv4Address(obj)

    def _encode(self, obj, context, path):
        return bytes(int(i) for i in obj.split("."))


class AliasedContainer(Container):
    ALIAS = None
    ORIGINAL = None

    def __getattr__(self, name):
        if name == self.ORIGINAL:
            name = self.ALIAS

        return super().__getattr__(name)

    def __getitem__(self, name):
        if name == self.ORIGINAL:
            name = self.ALIAS

        return super().__getitem__(name)


class EnumSwitch(Switch):
    def _emitparse(self, code):
        fname = "factory_%s" % code.allocateId()
        code.append(
            "%s = {%s}"
            % (
                fname,
                ", ".join(
                    "%r : lambda io,this: %s" % (int(key), sc._compileparse(code))
                    for key, sc in self.cases.items()
                ),
            )
        )

        defaultfname = "compiled_%s" % code.allocateId()
        code.append(
            "%s = lambda io,this: %s"
            % (
                defaultfname,
                self.default._compileparse(code),
            )
        )
        return "%s.get(%s, %s)(io, this)" % (fname, self.keyfunc, defaultfname)


class EnumSwitchStruct(Adapter):
    def __init__(self, subcon):
        assert isinstance(subcon, EnumSwitch), "subcon must be an EnumSwitch"
        assert subcon.default is Pass, "subcon must have a default Pass"

        super().__init__(subcon)

    def _decode(self, obj, context, path):
        keyfunc = self.subcon.keyfunc
        if callable(keyfunc):
            keyfunc = keyfunc(context)

        assert isinstance(keyfunc, enum.Enum), "keyfunc must be an enum"

        return {keyfunc.name.lower(): obj}

    def _encode(self, obj, context, path):
        keyfunc = self.subcon.keyfunc
        if callable(keyfunc):
            keyfunc = keyfunc(context)

        mapping = {_enum.value: _enum.name.lower() for _enum in self.subcon.cases.keys()}
        key = mapping[keyfunc]

        return obj[key]


enum_switch_struct_len_ = FuncPath(lambda arg: len(next(iter(arg.values()))))


class SwitchStruct(Adapter):
    def __init__(self, key, switch):
        super().__init__(Struct(key, switch))
        self.key = key
        self.switch = switch
        self._subcon = Struct(key, switch.subcon)

    def _decode(self, obj, context, path):
        key = self.switch.keyfunc(obj)

        try:
            value = obj[self.switch.name]
        except KeyError:
            value = obj[key.name.lower()]

        class _Container(AliasedContainer):
            ORIGINAL = self.switch.name
            ALIAS = key.name.lower()

        return _Container({self.key.name: key, key.name.lower(): value})

    def _encode(self, obj, context, path):
        keytype = self.key.subcon.type
        key = keytype(self.switch.keyfunc(obj))

        try:
            value = obj[key.name.lower()]
        except KeyError:
            value = obj[self.switch.name]

        return Container({self.key.name: key, self.switch.name: value})

    def _emitparse(self, code):
        keytype = self.key.subcon.type

        fname = f"parse_struct_{code.allocateId()}"
        block = f"""
            def {fname}(io, this):
                from {keytype.__module__} import {keytype.__name__}

                key = {self.key._compileparse(code)}
                key_name = {keytype.__name__}(key).name.lower()

                result = Container()
                this = Container(_ = this, _params = this['_params'], _root = None, _parsing = True, _building = False, _sizing = False, _subcons = None, _io = io, _index = this.get('_index', None))
                this['_root'] = this['_'].get('_root', this)
                try:
                    result[{self.key.name!r}] = this[{self.key.name!r}] = key
                    result[key_name] = this[key_name] = {self.switch._compileparse(code)}
                except StopFieldError:
                    pass
                return result
        """
        code.append(block)
        return f"{fname}(io, this)"


class NameAdapter(Adapter):
    def _decode(self, obj, context, path):
        obj._name = self.subcon.name
        return obj

    def _encode(self, obj, context, path):
        return obj.get(self.subcon.name, obj)


class NamedSelect(Adapter):
    def __init__(self, **subconskw):
        subcons = list(NameAdapter(k / v.compile()) for k, v in subconskw.items())
        super().__init__(Select(*subcons))
        self.__construct_doc__ = self._subcon = Select(**subconskw)

    def _decode(self, obj, context, path):
        return obj

    def _encode(self, obj, context, path):
        return obj


class IfThenElseDefault(IfThenElse):
    def __init__(self, condfunc, thensubcon, default):
        super().__init__(condfunc, thensubcon, Pass)
        self.condfunc = condfunc
        self.thensubcon = thensubcon
        self.default = default

    def _parse(self, stream, context, path):
        condfunc = self.condfunc
        if callable(condfunc):
            condfunc = condfunc(context)
        return self.thensubcon._parsereport(stream, context, path) if condfunc else self.default


def camelcase(field_name):
    if field_name is None:
        return None

    head, *tail = str(field_name).lower().replace(" ", "_").split("_")
    return "".join([head, *(i.title() for i in tail)])


def snakecase(camel_input):
    words = re.findall(r"[A-Z]?[a-z\d]+|[A-Z]{1,}(?=[A-Z][a-z]|\d|\W|$)|\d+", camel_input)
    return "_".join(map(str.lower, words))


def to_case_dict(value, case):
    if isinstance(value, dict):
        new_dict = {case(k): to_case_dict(v, case) for k, v in value.items() if not k.startswith("_")}

        name = getattr(value, "_name", None)
        return {name: new_dict} if name else new_dict

    elif isinstance(value, (set, list)):
        return [to_case_dict(i, case) for i in value]

    elif isinstance(value, enum.Enum):
        return value.value

    elif isinstance(value, bytes):
        return value.hex()

    if isinstance(value, datetime):
        return dict(
            year=value.year,
            month=value.month,
            day=value.day,
            hour=value.hour,
            minute=value.minute,
            second=value.second,
            microsecond=value.microsecond,
            timeZoneOffset=(value.utcoffset().total_seconds() / 60),
        )

    if isinstance(value, date):
        return dict(year=value.year, month=value.month, day=value.day)

    if isinstance(value, timedelta):
        return value.total_seconds()

    return value


def to_camelcase_dict(value):
    return to_case_dict(value, case=camelcase)


def to_snakecase_dict(value):
    return to_case_dict(value, case=snakecase)



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import pytest

from bluetooth_mesh.messages.config import StatusCode
from bluetooth_mesh.messages.light import (
    LightCTLMessage,
    LightCTLOpcode,
    LightCTLSetupMessage,
    LightCTLSetupOpcode,
    LightLightnessMessage,
    LightLightnessOpcode,
    LightLightnessSetupMessage,
    LightLightnessSetupOpcode,
)

valid_lightness = [
    # fmt: off
    pytest.param(
        b'\x82\x4b',
        LightLightnessOpcode.LIGHT_LIGHTNESS_GET,
        dict(),
        id="LIGHTNESS_GET"),
    pytest.param(
        b'\x82\x4c\xbb\xaa\x22',
        LightLightnessOpcode.LIGHT_LIGHTNESS_SET,
        dict(lightness=0xaabb,
             tid=34),
        id="LIGHTNESS_SET"),
    pytest.param(
        b'\x82\x4c\x01\x00\x22',
        LightLightnessOpcode.LIGHT_LIGHTNESS_SET,
        dict(lightness=1,
             tid=34,
             transition_time=6.3,
             delay=0.3),
        id="LIGHTNESS_SET_invalid"),
    pytest.param(
        b'\x82\x4c\x00\x00\x31\x32\x3c',
        LightLightnessOpcode.LIGHT_LIGHTNESS_SET,
        dict(lightness=0,
             tid=49,
             transition_time=5,
             delay=0.3),
        id="LIGHTNESS_SET_with_optional"),
    pytest.param(
        b'\x82\x4d\x00\x00\x31\x32\x3c',
        LightLightnessOpcode.LIGHT_LIGHTNESS_SET_UNACKNOWLEDGED,
        dict(lightness=0,
             tid=49,
             transition_time=5,
             delay=0.3),
        id="LIGHTNESS_SET_UNACKNOWLEDGED"),
    pytest.param(
        b'\x82\x4e\x44\x00',
        LightLightnessOpcode.LIGHT_LIGHTNESS_STATUS,
        dict(present_lightness=0x44),
        id="LIGHTNESS_STATUS"),
    pytest.param(
        b'\x82\x4e\x00\x00\x31\xc8\x0f',
        LightLightnessOpcode.LIGHT_LIGHTNESS_STATUS,
        dict(present_lightness=0,
             target_lightness=0xc831,
             remaining_time=1.5),
        id="LIGHTNESS_STATUS_with_optional"),

    pytest.param(
        b'\x82\x4f',
        LightLightnessOpcode.LIGHT_LIGHTNESS_LINEAR_GET,
        dict(),
        id="LIGHTNESS_LINEAR_GET"),
    pytest.param(
        b'\x82\x50\xbb\xaa\x01',
        LightLightnessOpcode.LIGHT_LIGHTNESS_LINEAR_SET,
        dict(lightness=0xaabb,
             tid=1),
        id="LIGHTNESS_LINEAR_SET"),
    pytest.param(
        b'\x82\x50\x01\x00\x22',
        LightLightnessOpcode.LIGHT_LIGHTNESS_LINEAR_SET,
        dict(lightness=1,
             tid=34,
             transition_time=6.3,
             delay=0.3),
        id="LIGHTNESS_LINEAR_SET_invalid"),
    pytest.param(
        b'\x82\x50\x00\x00\x31\x32\x1a',
        LightLightnessOpcode.LIGHT_LIGHTNESS_LINEAR_SET,
        dict(lightness=0,
             tid=49,
             transition_time=5,
             delay=0.13),
        id="LIGHTNESS_LINEAR_SET_with_optional"),
    pytest.param(
        b'\x82\x51\xff\x00\x31\x32\x3c',
        LightLightnessOpcode.LIGHT_LIGHTNESS_LINEAR_SET_UNACKNOWLEDGED,
        dict(lightness=0xff,
             tid=49,
             transition_time=5,
             delay=0.3),
        id="LIGHTNESS_LINEAR_SET_UNACKNOWLEDGED"),
    pytest.param(
        b'\x82\x52\x00\x00\xdd\xbb\x4c',
        LightLightnessOpcode.LIGHT_LIGHTNESS_LINEAR_STATUS,
        dict(present_lightness=0,
             target_lightness=0xbbdd,
             remaining_time=12),
        id="LIGHTNESS_LINEAR_STATUS"),

    pytest.param(
        b'\x82\x53',
        LightLightnessOpcode.LIGHT_LIGHTNESS_LAST_GET,
        dict(),
        id="LIGHTNESS_LAST_GET"),
    pytest.param(
        b'\x82\x54\x00\x00',
        LightLightnessOpcode.LIGHT_LIGHTNESS_LAST_STATUS,
        dict(lightness=0),
        id="LIGHTNESS_LAST_STATUS"),

    pytest.param(
        b'\x82\x55',
        LightLightnessOpcode.LIGHT_LIGHTNESS_DEFAULT_GET,
        dict(),
        id="LIGHTNESS_DEFAULT_GET"),
    pytest.param(
        b'\x82\x56\x00\x00',
        LightLightnessOpcode.LIGHT_LIGHTNESS_DEFAULT_STATUS,
        dict(lightness=0),
        id="LIGHTNESS_STATUS"),

    pytest.param(
        b'\x82\x57',
        LightLightnessOpcode.LIGHT_LIGHTNESS_RANGE_GET,
        dict(),
        id="LIGHTNESS_RANGE_GET"),
    pytest.param(
        b'\x82\x58\x00\x11\x11\x88\x88',
        LightLightnessOpcode.LIGHT_LIGHTNESS_RANGE_STATUS,
        dict(status=StatusCode.SUCCESS,
             range_min=0x1111,
             range_max=0x8888),
        id="LIGHTNESS_RANGE_STATUS"),
    # fmt: on
]

valid_lightness_setup = [
    # fmt: off
    pytest.param(
        b'\x82\x59\xbb\xaa',
        LightLightnessSetupOpcode.LIGHT_LIGHTNESS_SETUP_DEFAULT_SET,
        dict(lightness=0xaabb),
        id="LIGHTNESS_DEFAULT_SET"),
    pytest.param(
        b'\x82\x59\x01\x00',
        LightLightnessSetupOpcode.LIGHT_LIGHTNESS_SETUP_DEFAULT_SET,
        dict(lightness=1,
             tid=34,
             transition_time=6.3,
             delay=0.3),
        id="LIGHTNESS_DEFAULT_SET_invalid"),
    pytest.param(
        b'\x82\x5a\x00\x00',
        LightLightnessSetupOpcode.LIGHT_LIGHTNESS_SETUP_DEFAULT_SET_UNACKNOWLEDGED,
        dict(lightness=0),
        id="LIGHTNESS_SET_UNACKNOWLEDGED"),

    pytest.param(
        b'\x82\x5b\x11\x00\xff\xee',
        LightLightnessSetupOpcode.LIGHT_LIGHTNESS_SETUP_RANGE_SET,
        dict(range_min=0x0011,
             range_max=0xeeff),
        id="LIGHTNESS_RANGE_SET"),
    pytest.param(
        b'\x82\x5c\xcd\xab\x34\x12',
        LightLightnessSetupOpcode.LIGHT_LIGHTNESS_SETUP_RANGE_SET_UNACKNOWLEDGED,
        dict(range_min=0xabcd,
             range_max=0x1234),
        id="LIGHTNESS_RANGE_SET_UNACKNOWLEDGED"),
    # fmt: on
]


valid_ctl = [
    # fmt: off
    pytest.param(
        b'\x82\x5d',
        LightCTLOpcode.LIGHT_CTL_GET,
        dict(),
        id="CTL_GET"),
    pytest.param(
        b'\x82\x5e\xbb\xaa\xdd\xcc\xff\xee\x01',
        LightCTLOpcode.LIGHT_CTL_SET,
        dict(ctl_lightness=0xaabb,
             ctl_temperature=0xccdd,
             ctl_delta_uv=0xeeff,
             tid=1),
        id="CTL_SET"),
    pytest.param(
        b'\x82\x5e\xbb\xaa\xdd\xcc\xff\xee\x22',
        LightCTLOpcode.LIGHT_CTL_SET,
        dict(ctl_lightness=0xaabb,
             ctl_temperature=0xccdd,
             ctl_delta_uv=0xeeff,
             tid=34,
             transition_time=6.3,
             delay=0.3),
        id="CTL_SET_invalid"),
    pytest.param(
        b'\x82\x5e\xbb\xaa\xdd\xcc\xff\xee\x22\x32\x3c',
        LightCTLOpcode.LIGHT_CTL_SET,
        dict(ctl_lightness=0xaabb,
             ctl_temperature=0xccdd,
             ctl_delta_uv=0xeeff,
             tid=34,
             transition_time=5,
             delay=0.3),
        id="CTL_SET_with_optional"),
    pytest.param(
        b'\x82\x5f\x11\x11\x22\x22\x33\x33\x22\x32\x3c',
        LightCTLOpcode.LIGHT_CTL_SET_UNACKNOWLEDGED,
        dict(ctl_lightness=0x1111,
             ctl_temperature=0x2222,
             ctl_delta_uv=0x3333,
             tid=34,
             transition_time=5,
             delay=0.3),
        id="CTL_SET_UNACKNOWLEDGED"),
    pytest.param(
        b'\x82\x60\x11\x11\x22\x22',
        LightCTLOpcode.LIGHT_CTL_STATUS,
        dict(present_ctl_lightness=0x1111,
             present_ctl_temperature=0x2222),
        id="CTL_STATUS"),
    pytest.param(
        b'\x82\x60\x11\x11\x22\x22\x33\x33\x44\x44\x0f',
        LightCTLOpcode.LIGHT_CTL_STATUS,
        dict(present_ctl_lightness=0x1111,
             present_ctl_temperature=0x2222,
             target_ctl_lightness=0x3333,
             target_ctl_temperature=0x4444,
             remaining_time=1.5),
        id="CTL_STATUS_with_optional"),

    pytest.param(
        b'\x82\x61',
        LightCTLOpcode.LIGHT_CTL_TEMPERATURE_GET,
        dict(),
        id="CTL_TEMPERATURE_GET"),
    pytest.param(
        b'\x82\x64\xdd\xcc\xff\xee\x0a',
        LightCTLOpcode.LIGHT_CTL_TEMPERATURE_SET,
        dict(ctl_temperature=0xccdd,
             ctl_delta_uv=0xeeff,
             tid=10),
        id="CTL_TEMPERATURE_SET"),
    pytest.param(
        b'\x82\x64\xdd\xcc\xff\xee\x22\x32\x14',
        LightCTLOpcode.LIGHT_CTL_TEMPERATURE_SET,
        dict(ctl_temperature=0xccdd,
             ctl_delta_uv=0xeeff,
             tid=34,
             transition_time=5,
             delay=0.1),
        id="CTL_TEMPERATURE_SET_with_optional"),
    pytest.param(
        b'\x82\x65\x22\x22\x33\x33\xff\x32\x3c',
        LightCTLOpcode.LIGHT_CTL_TEMPERATURE_SET_UNACKNOWLEDGED,
        dict(ctl_temperature=0x2222,
             ctl_delta_uv=0x3333,
             tid=0xff,
             transition_time=5,
             delay=0.3),
        id="CTL_TEMPERATURE_SET_UNACKNOWLEDGED"),
    pytest.param(
        b'\x82\x66\x22\x22\xff\xff',
        LightCTLOpcode.LIGHT_CTL_TEMPERATURE_STATUS,
        dict(present_ctl_temperature=0x2222,
             present_ctl_delta_uv=0xffff),
        id="CTL_TEMPERATURE_STATUS"),
    pytest.param(
        b'\x82\x66\x11\x11\x22\x22\x33\x33\x44\x44\x0f',
        LightCTLOpcode.LIGHT_CTL_TEMPERATURE_STATUS,
        dict(present_ctl_temperature=0x1111,
             present_ctl_delta_uv=0x2222,
             target_ctl_temperature=0x3333,
             target_ctl_delta_uv=0x4444,
             remaining_time=1.5),
        id="CTL_TEMPERATURE_STATUS_with_optional"),

    pytest.param(
        b'\x82\x62',
        LightCTLOpcode.LIGHT_CTL_TEMPERATURE_RANGE_GET,
        dict(),
        id="CTL_TEMPERATURE_RANGE_GET"),
    pytest.param(
        b'\x82\x63\x00\xdd\xcc\xff\xee',
        LightCTLOpcode.LIGHT_CTL_TEMPERATURE_RANGE_STATUS,
        dict(status=StatusCode.SUCCESS,
             range_min=0xccdd,
             range_max=0xeeff),
        id="CTL_TEMPERATURE_RANGE_STATUS"),

    pytest.param(
        b'\x82\x67',
        LightCTLOpcode.LIGHT_CTL_TEMPERATURE_DEFAULT_GET,
        dict(),
        id="CTL_TEMPERATURE_DEFAULT_GET"),
    pytest.param(
        b'\x82\x68\xdd\xcc\xff\xee\x11\x11',
        LightCTLOpcode.LIGHT_CTL_TEMPERATURE_DEFAULT_STATUS,
        dict(ctl_lightness=0xccdd,
             ctl_temperature=0xeeff,
             ctl_delta_uv=0x1111),
        id="CTL_TEMPERATURE_DEFAULT_STATUS"),
    # fmt: on
]

valid_ctl_setup = [
    # fmt: off
    pytest.param(
        b'\x82\x69\xdd\xcc\xff\xee\x11\x11',
        LightCTLSetupOpcode.LIGHT_CTL_SETUP_TEMPERATURE_DEFAULT_SET,
        dict(ctl_lightness=0xccdd,
             ctl_temperature=0xeeff,
             ctl_delta_uv=0x1111),
        id="CTL_TEMPERATURE_DEFAULT_SET"),
    pytest.param(
        b'\x82\x6a\xdd\xcc\xff\xee\x11\x11',
        LightCTLSetupOpcode.LIGHT_CTL_SETUP_TEMPERATURE_DEFAULT_SET_UNACKNOWLEDGED,
        dict(ctl_lightness=0xccdd,
             ctl_temperature=0xeeff,
             ctl_delta_uv=0x1111),
        id="CTL_TEMPERATURE_DEFAULT_SET_UNACKNOWLEDGED"),

    pytest.param(
        b'\x82\x6b\xdd\xcc\xff\xee',
        LightCTLSetupOpcode.LIGHT_CTL_SETUP_TEMPERATURE_RANGE_SET,
        dict(range_min=0xccdd,
             range_max=0xeeff),
        id="CTL_TEMPERATURE_RANGE_SET"),
    pytest.param(
        b'\x82\x6c\x11\x11\x22\x22',
        LightCTLSetupOpcode.LIGHT_CTL_SETUP_TEMPERATURE_RANGE_SET_UNACKNOWLEDGED,
        dict(range_min=0x1111,
             range_max=0x2222),
        id="CTL_TEMPERATURE_RANGE_SET_UNACKNOWLEDGED"),
    # fmt: on
]


@pytest.mark.parametrize("encoded,opcode,data", valid_lightness)
def test_parse_valid_lightness(encoded, opcode, data):
    assert LightLightnessMessage.parse(encoded).params == data


@pytest.mark.parametrize("encoded,opcode,data", valid_lightness)
def test_build_valid_lightness(encoded, opcode, data):
    assert LightLightnessMessage.build(dict(opcode=opcode, params=data)) == encoded


@pytest.mark.parametrize("encoded,opcode,data", valid_lightness_setup)
def test_parse_valid_lightness_setup(encoded, opcode, data):
    assert LightLightnessSetupMessage.parse(encoded).params == data


@pytest.mark.parametrize("encoded,opcode,data", valid_lightness_setup)
def test_build_valid_lightness_setup(encoded, opcode, data):
    assert LightLightnessSetupMessage.build(dict(opcode=opcode, params=data)) == encoded


@pytest.mark.parametrize("encoded,opcode,data", valid_ctl)
def test_parse_valid_ctl(encoded, opcode, data):
    assert LightCTLMessage.parse(encoded).params == data


@pytest.mark.parametrize("encoded,opcode,data", valid_ctl)
def test_build_valid_ctl(encoded, opcode, data):
    assert LightCTLMessage.build(dict(opcode=opcode, params=data)) == encoded


@pytest.mark.parametrize("encoded,opcode,data", valid_ctl_setup)
def test_parse_valid_ctl_setup(encoded, opcode, data):
    assert LightCTLSetupMessage.parse(encoded).params == data


@pytest.mark.parametrize("encoded,opcode,data", valid_ctl_setup)
def test_build_valid_ctl_setup(encoded, opcode, data):
    assert LightCTLSetupMessage.build(dict(opcode=opcode, params=data)) == encoded



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import pytest

from bluetooth_mesh.messages.silvair.debugV2 import (
    DebugV2DataTypes,
    DebugV2ParameterTypes,
    DebugV2SubMessage,
    DebugV2SubOpcode,
)

# fmt: off
valid = [
    pytest.param(b'\x02\x09\x00\x02\x00\x10\x20\x30', DebugV2SubOpcode.STATUS, DebugV2ParameterTypes.UPTIME, 0x00,
                 [{"data_type": DebugV2DataTypes.UINT32, "parameter": 807407616}], id="StatusUptime"),
    pytest.param(b'\x02\x08\x00\x04\x20', DebugV2SubOpcode.STATUS, DebugV2ParameterTypes.RSSI_THRESHOLD, 0x00,
                 [{"data_type": DebugV2DataTypes.INT8, "parameter": 0x20}], id="StatusRSSIThreshold"),
    pytest.param(b'\x02\x0B\x02\x0C\x06\x00\x10\x00\x20\x00\x30', DebugV2SubOpcode.STATUS,
                 DebugV2ParameterTypes.TIME_SYNC_MSG_LIST, 0x02,
                 [{"data_type": DebugV2DataTypes.ARRAY, "parameter": [0x00, 0x10, 0x00, 0x20, 0x00, 0x30]}],
                 id="StatusTimeSyncMsgList"),
    pytest.param(b'\x02\x4B\x00\x0B\x09my string', DebugV2SubOpcode.STATUS, 0x4B, 0x00,
                 [{"data_type": DebugV2DataTypes.STRING, "parameter": "my string"}], id="StatusUnknownParameterString"),
    pytest.param(b'\x02\x55\x00\x02\x00\x10\x20\x30\x0A\x00\x08\x00\x32\x00\x28', DebugV2SubOpcode.STATUS, 0x55, 0x00,
                 [{"data_type": DebugV2DataTypes.UINT32, "parameter": 807407616},
                  {"data_type": DebugV2DataTypes.ENUM, "parameter": 0},
                  {"data_type": DebugV2DataTypes.FLOAT, "parameter": 7.116269379325857e-15}],
                 id="StatusUnknownParameterMultipleFields"),
    pytest.param(b'\x03\x13\x00', DebugV2SubOpcode.CLEAR, DebugV2ParameterTypes.OCCUPANCY_HISTORY, 0x00, [],
                 id="ClearMsgWithoutAnyPayload"),
    pytest.param(b'\x00\x0F\x00', DebugV2SubOpcode.GET, DebugV2ParameterTypes.RECALLED_EVENTS_TIME, 0x00, [],
                 id="GetMsgWithoutAnyPayload"),
    pytest.param(b'\x02\xFE\x00\xFF', DebugV2SubOpcode.STATUS, 0xFE, 0x00,
                 [{"data_type": DebugV2DataTypes.INVALID, "parameter": {}}],
                 id="StatusMsgWithInvalidDataType"),
]


# fmt: on


@pytest.mark.parametrize("encoded, subopcode, parameter_type, page_number, payload", valid)
def test_parse_valid(encoded, subopcode, parameter_type, page_number, payload):
    decoded = DebugV2SubMessage.parse(encoded)
    assert decoded.subopcode == subopcode
    assert decoded.parameter_type == parameter_type
    assert decoded.page_number == page_number
    for idx, element in enumerate(decoded.payload):
        assert element.data_type == payload[idx]["data_type"]
        assert element.parameter == payload[idx]["parameter"]


@pytest.mark.parametrize("encoded, subopcode, parameter_type, page_number, payload", valid)
def test_build_valid(encoded, subopcode, parameter_type, page_number, payload):
    assert (
        DebugV2SubMessage.build(
            dict(
                subopcode=subopcode,
                parameter_type=parameter_type,
                page_number=page_number,
                payload=payload,
            )
        )
        == encoded
    )



import tempfile

import capnp

from .generator import generate


def load_definitions():
    with tempfile.NamedTemporaryFile(suffix=".capnp", mode="w+", encoding="utf-8") as tmp_file:
        generate(0x97EBBC55406EDFB4, file=tmp_file)
        tmp_file.seek(0)
        source = tmp_file.read()

        capnp.remove_import_hook()
        messages = capnp.load(tmp_file.name)

        return source, messages


SOURCE, MESSAGES = load_definitions()

AccessMessage = MESSAGES.AccessMessage
